import sys
import warnings
from typing import Dict, List, Optional, Tuple, Union

from packaging.version import Version

from snowflake.ml._internal import telemetry
from snowflake.snowpark import AsyncJob, Row, Session
from snowflake.snowpark._internal import utils as snowpark_utils

cache: Dict[str, Optional[str]] = {}

_PROJECT = "ModelDevelopment"
_SUBPROJECT = "utils"

_RUNTIME_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"

_relax_version: bool = False


def is_relaxed() -> bool:
    return _relax_version


def get_valid_pkg_versions_supported_in_snowflake_conda_channel(
    pkg_versions: List[str], session: Session, subproject: Optional[str] = None
) -> List[str]:
    if snowpark_utils.is_in_stored_procedure():  # type: ignore[no-untyped-call]
        return _get_valid_pkg_versions_supported_in_snowflake_conda_channel_sync(pkg_versions, session, subproject)
    else:
        return _get_valid_pkg_versions_supported_in_snowflake_conda_channel_async(pkg_versions, session, subproject)


def _get_valid_pkg_versions_supported_in_snowflake_conda_channel_sync(
    pkg_versions: List[str], session: Session, subproject: Optional[str] = None
) -> List[str]:
    for pkg_version in pkg_versions:
        if pkg_version not in cache:
            pkg_version_list = _query_pkg_version_supported_in_snowflake_conda_channel(
                pkg_version=pkg_version, session=session, block=True, subproject=subproject
            )
            assert isinstance(pkg_version_list, list)  # keep mypy happy
            try:
                cache[pkg_version] = pkg_version_list[0]["VERSION"]
            except IndexError:
                cache[pkg_version] = None

    pkg_version_conda_list = _get_conda_packages_and_emit_warnings(pkg_versions)

    return pkg_version_conda_list


def _get_valid_pkg_versions_supported_in_snowflake_conda_channel_async(
    pkg_versions: List[str], session: Session, subproject: Optional[str] = None
) -> List[str]:
    pkg_version_async_job_list: List[Tuple[str, AsyncJob]] = []
    for pkg_version in pkg_versions:
        if pkg_version not in cache:
            # Execute pkg version queries asynchronously.
            async_job = _query_pkg_version_supported_in_snowflake_conda_channel(
                pkg_version=pkg_version, session=session, block=False, subproject=subproject
            )
            assert isinstance(async_job, AsyncJob)
            pkg_version_async_job_list.append((pkg_version, async_job))

    # Populate the cache.
    for pkg_version, pkg_version_async_job in pkg_version_async_job_list:
        pkg_version_list = pkg_version_async_job.result()
        assert isinstance(pkg_version_list, list)
        try:
            cache[pkg_version] = pkg_version_list[0]["VERSION"]
        except IndexError:
            cache[pkg_version] = None

    pkg_version_conda_list = _get_conda_packages_and_emit_warnings(pkg_versions)

    return pkg_version_conda_list


def _query_pkg_version_supported_in_snowflake_conda_channel(
    pkg_version: str, session: Session, block: bool, subproject: Optional[str] = None
) -> Union[AsyncJob, List[Row]]:
    tokens = pkg_version.split("==")
    if len(tokens) != 2:
        raise RuntimeError(
            "Expected package name and versions to specified in format "
            f"'<pkg_name>==<version>', but found {pkg_version}"
        )
    pkg_name = tokens[0]
    version = Version(tokens[1])
    major_version, minor_version, micro_version = version.major, version.minor, version.micro

    # relax version control - only major_version.minor_version.* will be enforced.
    # the result would be ordered by, the version that closest to user's version, and the latest.
    sql = f"""
        SELECT PACKAGE_NAME, VERSION, LANGUAGE
        FROM (
            SELECT *,
            SUBSTRING(VERSION, LEN(VERSION) - CHARINDEX('.', REVERSE(VERSION)) + 2, LEN(VERSION)) as micro_version
            FROM information_schema.packages
            WHERE package_name = '{pkg_name}'
            AND version LIKE '{major_version}.{minor_version}.%'
            ORDER BY abs({micro_version}-micro_version), -micro_version
        )
        """
    result_df = session.sql(sql)

    # TODO(snandamuri): Move this filter into main SQL query after BCR 7.19 is completed.
    if "RUNTIME_VERSION" in result_df.columns:
        result_df = result_df.filter(f"RUNTIME_VERSION = {_RUNTIME_VERSION}")

    pkg_version_list_or_async_job = result_df.collect(
        statement_params=telemetry.get_statement_params(_PROJECT, subproject or _SUBPROJECT),
        block=block,
    )

    return pkg_version_list_or_async_job


def _get_conda_packages_and_emit_warnings(pkg_versions: List[str]) -> List[str]:
    pkg_version_conda_list: List[str] = []
    pkg_version_warning_list: List[List[str]] = []
    for pkg_version in pkg_versions:
        try:
            conda_pkg_version = cache[pkg_version]
        except KeyError:
            conda_pkg_version = None

        if conda_pkg_version is None:
            if _relax_version:
                pkg_version_warning_list.append([pkg_version, _RUNTIME_VERSION])
            else:
                raise RuntimeError(
                    f"Package {pkg_version} is not supported in snowflake conda channel for "
                    f"python runtime {_RUNTIME_VERSION}."
                )
        else:
            tokens = pkg_version.split("==")
            pkg_name = tokens[0]
            pkg_version_conda_list.append(f"{pkg_name}=={conda_pkg_version}")

    if pkg_version_warning_list:
        warnings.warn(
            f"Package {', '.join([pkg[0] for pkg in pkg_version_warning_list])} is not supported "
            f"in snowflake conda channel for python runtime "
            f"{', '.join([pkg[1] for pkg in pkg_version_warning_list])}."
        )

    return pkg_version_conda_list



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#
import sys
from io import BytesIO
from typing import Any

import cloudpickle

import snowflake.snowpark._internal.utils as snowpark_utils
from snowflake import snowpark

_RESULT_SIZE_THRESHOLD = 5 * (1024**2)  # 5MB


# This module handles serialization, uploading, downloading, and deserialization of stored
# procedure results. If the results are too large to be returned from a stored procedure,
# the result will be uploaded. The client can then retrieve and deserialize the result if
# it was uploaded.


def serialize(session: snowpark.Session, result: Any) -> bytes:
    """
    Serialize a tuple containing the result (or None) and the result object filepath
    if the result was uploaded to a stage (or None).

    Args:
        session: Snowpark session.
        result: Object to be serialized.

    Returns:
        Cloudpickled string of bytes of the result tuple.
    """
    result_object_filepath = None
    result_bytes = cloudpickle.dumps(result)
    if sys.getsizeof(result_bytes) > _RESULT_SIZE_THRESHOLD:
        stage_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.STAGE)
        session.sql(f"CREATE TEMPORARY STAGE {stage_name}").collect()
        result_object_filepath = f"@{stage_name}/{snowpark_utils.generate_random_alphanumeric()}"
        session.file.put_stream(BytesIO(result_bytes), result_object_filepath)
        result_object_filepath = f"{result_object_filepath}.gz"

    if result_object_filepath is not None:
        return cloudpickle.dumps((None, result_object_filepath))  # type: ignore[no-any-return]

    return cloudpickle.dumps((result, None))  # type: ignore[no-any-return]


def deserialize(session: snowpark.Session, result_bytes: bytes) -> Any:
    """
    Loads and/or deserializes the (maybe uploaded) result.

    Args:
        session: Snowpark session.
        result_bytes: String of bytes returned by serialize method.

    Returns:
        The deserialized result (any type).
    """
    result_object, result_object_filepath = cloudpickle.loads(result_bytes)
    if result_object_filepath is not None:
        result_object_bytes_io = session.file.get_stream(result_object_filepath, decompress=True)
        result_bytes = result_object_bytes_io.read()
        result_object = cloudpickle.loads(result_bytes)

    return result_object



import copy
import pathlib
from typing import List, Optional

from packaging import requirements

from snowflake.ml._internal import env as snowml_env, env_utils, file_utils
from snowflake.ml.model._model_composer.model_manifest import model_manifest_schema
from snowflake.ml.model._model_composer.model_runtime import _runtime_requirements
from snowflake.ml.model._packager.model_env import model_env
from snowflake.ml.model._packager.model_meta import model_meta as model_meta_api
from snowflake.snowpark import session

_UDF_INFERENCE_DEPENDENCIES = _runtime_requirements.REQUIREMENTS


class ModelRuntime:
    """Class to represent runtime in a model, which controls the runtime and version, imports and dependencies.

    Attributes:
        model_meta: Model Metadata.
        runtime_env: ModelEnv object representing the actual environment when deploying. The environment is based on
            the environment from the packaged model with additional dependencies required to deploy.
        imports: List of files to be imported in the created functions. At least packed model should be imported.
            If the required Snowpark ML library is not available in the server-side, we will automatically pack the
            local version as well as "snowflake-ml-python.zip" and added into the imports.
    """

    RUNTIME_DIR_REL_PATH = "runtimes"

    def __init__(
        self,
        session: session.Session,
        name: str,
        model_meta: model_meta_api.ModelMetadata,
        imports: Optional[List[pathlib.PurePosixPath]] = None,
    ) -> None:
        self.name = name
        self.model_meta = model_meta
        self.runtime_env = copy.deepcopy(self.model_meta.env)
        self.imports = imports or []

        snowml_pkg_spec = f"{env_utils.SNOWPARK_ML_PKG_NAME}=={self.runtime_env.snowpark_ml_version}"
        if self.runtime_env._snowpark_ml_version.local:
            self.embed_local_ml_library = True
        else:
            snowml_server_availability = env_utils.validate_requirements_in_snowflake_conda_channel(
                session=session,
                reqs=[requirements.Requirement(snowml_pkg_spec)],
                python_version=snowml_env.PYTHON_VERSION,
            )
            self.embed_local_ml_library = snowml_server_availability is None

        if self.embed_local_ml_library:
            self.runtime_env.include_if_absent(
                [
                    model_env.ModelDependency(requirement=dep, pip_name=requirements.Requirement(dep).name)
                    for dep in _UDF_INFERENCE_DEPENDENCIES
                ],
                check_local_version=True,
            )
        else:
            self.runtime_env.include_if_absent(
                [
                    model_env.ModelDependency(requirement=dep, pip_name=requirements.Requirement(dep).name)
                    for dep in _UDF_INFERENCE_DEPENDENCIES + [snowml_pkg_spec]
                ],
                check_local_version=True,
            )

    def save(self, workspace_path: pathlib.Path) -> model_manifest_schema.ModelRuntimeDict:
        runtime_base_path = workspace_path / ModelRuntime.RUNTIME_DIR_REL_PATH / self.name
        runtime_base_path.mkdir(parents=True, exist_ok=True)

        if self.embed_local_ml_library:
            snowpark_ml_lib_path = runtime_base_path / "snowflake-ml-python.zip"
            file_utils.zip_python_package(str(snowpark_ml_lib_path), "snowflake.ml")
            snowpark_ml_lib_rel_path = pathlib.PurePosixPath(
                snowpark_ml_lib_path.relative_to(workspace_path).as_posix()
            )
            self.imports.append(snowpark_ml_lib_rel_path)

        env_dict = self.runtime_env.save_as_dict(runtime_base_path)
        return model_manifest_schema.ModelRuntimeDict(
            language="PYTHON",
            version=self.runtime_env.python_version,
            imports=list(map(str, self.imports)),
            dependencies=model_manifest_schema.ModelRuntimeDependenciesDict(
                conda=(runtime_base_path / env_dict["conda"]).relative_to(workspace_path).as_posix()
            ),
        )



from typing import List, Optional, Tuple, Union

import cloudpickle
import numpy as np
import numpy.typing as npt
import sklearn
from packaging import version
from sklearn import metrics

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.utils import result
from snowflake.ml.modeling.metrics import metrics_utils
from snowflake.snowpark import functions as F
from snowflake.snowpark._internal import utils as snowpark_utils

_PROJECT = "ModelDevelopment"
_SUBPROJECT = "Metrics"


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def precision_recall_curve(
    *,
    df: snowpark.DataFrame,
    y_true_col_name: str,
    probas_pred_col_name: str,
    pos_label: Optional[Union[str, int]] = None,
    sample_weight_col_name: Optional[str] = None,
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """
    Compute precision-recall pairs for different probability thresholds.

    Note: this implementation is restricted to the binary classification task.

    The precision is the ratio ``tp / (tp + fp)`` where ``tp`` is the number of
    true positives and ``fp`` the number of false positives. The precision is
    intuitively the ability of the classifier not to label as positive a sample
    that is negative.

    The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
    true positives and ``fn`` the number of false negatives. The recall is
    intuitively the ability of the classifier to find all the positive samples.

    The last precision and recall values are 1. and 0. respectively and do not
    have a corresponding threshold. This ensures that the graph starts on the
    y axis.

    The first precision and recall values are precision=class balance and recall=1.0
    which corresponds to a classifier that always predicts the positive class.

    Args:
        df: Input dataframe.
        y_true_col_name: Column name representing true binary labels.
            If labels are not either {-1, 1} or {0, 1}, then pos_label should be
            explicitly given.
        probas_pred_col_name: Column name representing target scores.
            Can either be probability estimates of the positive
            class, or non-thresholded measure of decisions (as returned by
            `decision_function` on some classifiers).
        pos_label: The label of the positive class.
            When ``pos_label=None``, if y_true is in {-1, 1} or {0, 1},
            ``pos_label`` is set to 1, otherwise an error will be raised.
        sample_weight_col_name: Column name representing sample weights.

    Returns:
        Tuple containing following items
            precision - ndarray of shape (n_thresholds + 1,)
                Precision values such that element i is the precision of
                predictions with score >= thresholds[i] and the last element is 1.
            recall - ndarray of shape (n_thresholds + 1,)
                Decreasing recall values such that element i is the recall of
                predictions with score >= thresholds[i] and the last element is 0.
            thresholds - ndarray of shape (n_thresholds,)
                Increasing thresholds on the decision function used to compute
                precision and recall.
    """
    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_name, probas_pred_col_name, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_result_module = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def precision_recall_curve_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_name]
        probas_pred = df[probas_pred_col_name]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None
        precision, recall, thresholds = metrics.precision_recall_curve(
            y_true,
            probas_pred,
            pos_label=pos_label,
            sample_weight=sample_weight,
        )
        result_module = cloudpickle.loads(pickled_result_module)
        return result_module.serialize(session, (precision, recall, thresholds))  # type: ignore[no-any-return]

    result_object = result.deserialize(session, precision_recall_curve_anon_sproc(session))
    res: Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]] = result_object
    return res


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def roc_auc_score(
    *,
    df: snowpark.DataFrame,
    y_true_col_names: Union[str, List[str]],
    y_score_col_names: Union[str, List[str]],
    average: Optional[str] = "macro",
    sample_weight_col_name: Optional[str] = None,
    max_fpr: Optional[float] = None,
    multi_class: str = "raise",
    labels: Optional[npt.ArrayLike] = None,
) -> Union[float, npt.NDArray[np.float_]]:
    """
    Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    from prediction scores.

    Note: this implementation can be used with binary, multiclass and
    multilabel classification, but some restrictions apply.

    Args:
        df: Input dataframe.
        y_true_col_names: Column name(s) representing true labels or binary label indicators.
            The binary and multiclass cases expect labels with shape (n_samples,)
            while the multilabel case expects binary label indicators with shape
            (n_samples, n_classes).
        y_score_col_names: Column name(s) representing target scores.
            * In the binary case, it corresponds to an array of shape
              `(n_samples,)`. Both probability estimates and non-thresholded
              decision values can be provided. The probability estimates correspond
              to the **probability of the class with the greater label**.
              The decision values correspond to the output of
              `estimator.decision_function`.
            * In the multiclass case, it corresponds to an array of shape
              `(n_samples, n_classes)` of probability estimates provided by the
              `predict_proba` method. The probability estimates **must**
              sum to 1 across the possible classes. In addition, the order of the
              class scores must correspond to the order of ``labels``,
              if provided, or else to the numerical or lexicographical order of
              the labels in ``y_true``.
            * In the multilabel case, it corresponds to an array of shape
              `(n_samples, n_classes)`. Probability estimates are provided by the
              `predict_proba` method and the non-thresholded decision values by
              the `decision_function` method. The probability estimates correspond
              to the **probability of the class with the greater label for each
              output** of the classifier.
        average: {'micro', 'macro', 'samples', 'weighted'} or None, default='macro'
            If ``None``, the scores for each class are returned.
            Otherwise, this determines the type of averaging performed on the data.
            Note: multiclass ROC AUC currently only handles the 'macro' and
            'weighted' averages. For multiclass targets, `average=None` is only
            implemented for `multi_class='ovr'` and `average='micro'` is only
            implemented for `multi_class='ovr'`.
            ``'micro'``
                Calculate metrics globally by considering each element of the label
                indicator matrix as a label.
            ``'macro'``
                Calculate metrics for each label, and find their unweighted
                mean.  This does not take label imbalance into account.
            ``'weighted'``
                Calculate metrics for each label, and find their average, weighted
                by support (the number of true instances for each label).
            ``'samples'``
                Calculate metrics for each instance, and find their average.
            Will be ignored when ``y_true`` is binary.
        sample_weight_col_name: Column name representing sample weights.
        max_fpr: float > 0 and <= 1, default=None
            If not ``None``, the standardized partial AUC [2]_ over the range
            [0, max_fpr] is returned. For the multiclass case, ``max_fpr``,
            should be either equal to ``None`` or ``1.0`` as AUC ROC partial
            computation currently is not supported for multiclass.
        multi_class: {'raise', 'ovr', 'ovo'}, default='raise'
            Only used for multiclass targets. Determines the type of configuration
            to use. The default value raises an error, so either
            ``'ovr'`` or ``'ovo'`` must be passed explicitly.
            ``'ovr'``
                Stands for One-vs-rest. Computes the AUC of each class
                against the rest [3]_ [4]_. This
                treats the multiclass case in the same way as the multilabel case.
                Sensitive to class imbalance even when ``average == 'macro'``,
                because class imbalance affects the composition of each of the
                'rest' groupings.
            ``'ovo'``
                Stands for One-vs-one. Computes the average AUC of all
                possible pairwise combinations of classes [5]_.
                Insensitive to class imbalance when
                ``average == 'macro'``.
        labels: Only used for multiclass targets. List of labels that index the
            classes in ``y_score``. If ``None``, the numerical or lexicographical
            order of the labels in ``y_true`` is used.

    Returns:
        Area Under the Curve score.
    """
    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_names, y_score_col_names, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_result_module = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def roc_auc_score_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_names]
        y_score = df[y_score_col_names]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None
        auc = metrics.roc_auc_score(
            y_true,
            y_score,
            average=average,
            sample_weight=sample_weight,
            max_fpr=max_fpr,
            multi_class=multi_class,
            labels=labels,
        )
        result_module = cloudpickle.loads(pickled_result_module)
        return result_module.serialize(session, auc)  # type: ignore[no-any-return]

    result_object = result.deserialize(session, roc_auc_score_anon_sproc(session))
    auc: Union[float, npt.NDArray[np.float_]] = result_object
    return auc


@telemetry.send_api_usage_telemetry(project=_PROJECT, subproject=_SUBPROJECT)
def roc_curve(
    *,
    df: snowpark.DataFrame,
    y_true_col_name: str,
    y_score_col_name: str,
    pos_label: Optional[Union[str, int]] = None,
    sample_weight_col_name: Optional[str] = None,
    drop_intermediate: bool = True,
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """
    Compute Receiver operating characteristic (ROC).

    Note: this implementation is restricted to the binary classification task.

    Args:
        df: Input dataframe.
        y_true_col_name: Column name representing true binary labels.
            If labels are not either {-1, 1} or {0, 1}, then pos_label should be
            explicitly given.
        y_score_col_name: Column name representing target scores, can either
            be probability estimates of the positive class, confidence values,
            or non-thresholded measure of decisions (as returned by
            "decision_function" on some classifiers).
        pos_label: The label of the positive class.
            When ``pos_label=None``, if `y_true` is in {-1, 1} or {0, 1},
            ``pos_label`` is set to 1, otherwise an error will be raised.
        sample_weight_col_name: Column name representing sample weights.
        drop_intermediate: Whether to drop some suboptimal thresholds which would
            not appear on a plotted ROC curve. This is useful in order to create
            lighter ROC curves.

    Returns:
        Tuple containing following items
            fpr - ndarray of shape (>2,)
                Increasing false positive rates such that element i is the false
                positive rate of predictions with score >= `thresholds[i]`.
            tpr - ndarray of shape (>2,)
                Increasing true positive rates such that element `i` is the true
                positive rate of predictions with score >= `thresholds[i]`.
            thresholds - ndarray of shape = (n_thresholds,)
                Decreasing thresholds on the decision function used to compute
                fpr and tpr. `thresholds[0]` represents no instances being predicted
                and is arbitrarily set to `max(y_score) + 1`.
    """
    session = df._session
    assert session is not None
    sproc_name = snowpark_utils.random_name_for_temp_object(snowpark_utils.TempObjectType.PROCEDURE)
    sklearn_release = version.parse(sklearn.__version__).release
    statement_params = telemetry.get_statement_params(_PROJECT, _SUBPROJECT)

    cols = metrics_utils.flatten_cols([y_true_col_name, y_score_col_name, sample_weight_col_name])
    queries = df[cols].queries["queries"]

    pickled_result_module = cloudpickle.dumps(result)

    @F.sproc(  # type: ignore[misc]
        is_permanent=False,
        session=session,
        name=sproc_name,
        replace=True,
        packages=[
            "cloudpickle",
            f"scikit-learn=={sklearn_release[0]}.{sklearn_release[1]}.*",
            "snowflake-snowpark-python",
        ],
        statement_params=statement_params,
        anonymous=True,
    )
    def roc_curve_anon_sproc(session: snowpark.Session) -> bytes:
        for query in queries[:-1]:
            _ = session.sql(query).collect(statement_params=statement_params)
        sp_df = session.sql(queries[-1])
        df = sp_df.to_pandas(statement_params=statement_params)
        df.columns = sp_df.columns

        y_true = df[y_true_col_name]
        y_score = df[y_score_col_name]
        sample_weight = df[sample_weight_col_name] if sample_weight_col_name else None
        fpr, tpr, thresholds = metrics.roc_curve(
            y_true,
            y_score,
            pos_label=pos_label,
            sample_weight=sample_weight,
            drop_intermediate=drop_intermediate,
        )

        result_module = cloudpickle.loads(pickled_result_module)
        return result_module.serialize(session, (fpr, tpr, thresholds))  # type: ignore[no-any-return]

    result_object = result.deserialize(session, roc_curve_anon_sproc(session))
    res: Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_], npt.NDArray[np.float_]] = result_object

    return res



from collections import abc
from typing import List, Literal, Sequence

import numpy as np
import pandas as pd
from typing_extensions import TypeGuard

from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._signatures import base_handler, core


class NumpyArrayHandler(base_handler.BaseDataHandler[model_types._SupportedNumpyArray]):
    @staticmethod
    def can_handle(data: model_types.SupportedDataType) -> TypeGuard[model_types._SupportedNumpyArray]:
        return isinstance(data, np.ndarray)

    @staticmethod
    def count(data: model_types._SupportedNumpyArray) -> int:
        return data.shape[0]

    @staticmethod
    def truncate(data: model_types._SupportedNumpyArray) -> model_types._SupportedNumpyArray:
        return data[: min(NumpyArrayHandler.count(data), NumpyArrayHandler.SIG_INFER_ROWS_COUNT_LIMIT)]

    @staticmethod
    def validate(data: model_types._SupportedNumpyArray) -> None:
        if data.shape == (0,):
            # Empty array
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError("Data Validation Error: Empty data is found."),
            )

        if data.shape == ():
            # scalar
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError("Data Validation Error: Scalar data is found."),
            )

    @staticmethod
    def infer_signature(
        data: model_types._SupportedNumpyArray, role: Literal["input", "output"]
    ) -> Sequence[core.BaseFeatureSpec]:
        feature_prefix = f"{NumpyArrayHandler.FEATURE_PREFIX}_"
        dtype = core.DataType.from_numpy_type(data.dtype)
        role_prefix = (NumpyArrayHandler.INPUT_PREFIX if role == "input" else NumpyArrayHandler.OUTPUT_PREFIX) + "_"
        if len(data.shape) == 1:
            return [core.FeatureSpec(dtype=dtype, name=f"{role_prefix}{feature_prefix}0")]
        else:
            # For high-dimension array, 0-axis is for batch, 1-axis is for column, further more is details of columns.
            features = []
            n_cols = data.shape[1]
            ft_names = [f"{role_prefix}{feature_prefix}{i}" for i in range(n_cols)]
            for col_data, ft_name in zip(data[0], ft_names):
                if isinstance(col_data, np.ndarray):
                    ft_shape = np.shape(col_data)
                    features.append(core.FeatureSpec(dtype=dtype, name=ft_name, shape=ft_shape))
                else:
                    features.append(core.FeatureSpec(dtype=dtype, name=ft_name))
            return features

    @staticmethod
    def convert_to_df(data: model_types._SupportedNumpyArray, ensure_serializable: bool = True) -> pd.DataFrame:
        if len(data.shape) == 1:
            data = np.expand_dims(data, axis=1)
        n_cols = data.shape[1]
        if len(data.shape) == 2:
            return pd.DataFrame(data)
        else:
            n_rows = data.shape[0]
            if ensure_serializable:
                return pd.DataFrame(data={i: [data[k, i].tolist() for k in range(n_rows)] for i in range(n_cols)})
            return pd.DataFrame(data={i: [list(data[k, i]) for k in range(n_rows)] for i in range(n_cols)})


class SeqOfNumpyArrayHandler(base_handler.BaseDataHandler[Sequence[model_types._SupportedNumpyArray]]):
    @staticmethod
    def can_handle(data: model_types.SupportedDataType) -> TypeGuard[Sequence[model_types._SupportedNumpyArray]]:
        if not isinstance(data, abc.Sequence):
            return False
        if len(data) == 0:
            return False
        if isinstance(data[0], np.ndarray):
            return all(isinstance(data_col, np.ndarray) for data_col in data)
        return False

    @staticmethod
    def count(data: Sequence[model_types._SupportedNumpyArray]) -> int:
        return min(NumpyArrayHandler.count(data_col) for data_col in data)

    @staticmethod
    def truncate(data: Sequence[model_types._SupportedNumpyArray]) -> Sequence[model_types._SupportedNumpyArray]:
        return [
            data_col[: min(SeqOfNumpyArrayHandler.count(data), SeqOfNumpyArrayHandler.SIG_INFER_ROWS_COUNT_LIMIT)]
            for data_col in data
        ]

    @staticmethod
    def validate(data: Sequence[model_types._SupportedNumpyArray]) -> None:
        for data_col in data:
            NumpyArrayHandler.validate(data_col)

    @staticmethod
    def infer_signature(
        data: Sequence[model_types._SupportedNumpyArray], role: Literal["input", "output"]
    ) -> Sequence[core.BaseFeatureSpec]:
        feature_prefix = f"{SeqOfNumpyArrayHandler.FEATURE_PREFIX}_"
        features: List[core.BaseFeatureSpec] = []
        role_prefix = (
            SeqOfNumpyArrayHandler.INPUT_PREFIX if role == "input" else SeqOfNumpyArrayHandler.OUTPUT_PREFIX
        ) + "_"

        for i, data_col in enumerate(data):
            dtype = core.DataType.from_numpy_type(data_col.dtype)
            ft_name = f"{role_prefix}{feature_prefix}{i}"
            if len(data_col.shape) == 1:
                features.append(core.FeatureSpec(dtype=dtype, name=ft_name))
            else:
                ft_shape = tuple(data_col.shape[1:])
                features.append(core.FeatureSpec(dtype=dtype, name=ft_name, shape=ft_shape))
        return features

    @staticmethod
    def convert_to_df(
        data: Sequence[model_types._SupportedNumpyArray], ensure_serializable: bool = True
    ) -> pd.DataFrame:
        if ensure_serializable:
            return pd.DataFrame(data={i: data_col.tolist() for i, data_col in enumerate(data)})
        return pd.DataFrame(data={i: list(data_col) for i, data_col in enumerate(data)})



from typing import Any, Dict, Generator, List

import fsspec
import numpy.typing as npt
import pyarrow as pa
import pyarrow.parquet as pq
import tensorflow as tf

from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml.fileset import parquet_parser


def read_and_parse_parquet(
    files: List[str],
    filesystem: fsspec.AbstractFileSystem,
    batch_size: int,
    shuffle: bool,
    drop_last_batch: bool,
) -> tf.data.Dataset:
    """Creates a tf.data.Dataset that reads given parquet files into batched Tensors.

    Args:
        files: A list of input parquet file URIs to read and parse. The parquet files should
            have the same schema.
        filesystem: A fsspec/pyarrow file system that is used to open given file URIs.
        batch_size: Specifies the size of each batch that will be yield. It is preferred to
            set it to your training batch size, and avoid using dataset.{batch(),rebatch()} later.
        shuffle: Whether the data in the file will be shuffled. If set to be true, it will first randomly shuffle
            the order of files, and then shuflle the order of rows in each file. It is preferred
            to shuffle the data this way than dataset.unbatch().shuffle().rebatch().
        drop_last_batch: Whether the last batch of data should be dropped. If set to be true, then the last batch will
            get dropped if its size is smaller than the given batch_size.

    Returns:
        A tf.data.Dataset generates batched Tensors in a dict. The keys will be the column names in
        the parquet files.

    Raises:
        SnowflakeMLException: if `files` is empty.

    Example:
        >>> from snowflake.ml.fileset import sfcfs, tf_dataset
        >>> conn = snowflake.connector.connect(**connection_parameters)
        >>> fs = sfcfs.SFFileSystem(conn)
        >>> files = fs.ls(dir_path)
        >>> ds = tf_dataset.parse_and_read_parquet(files, fs, batch_size = 2)
        >>> for batch in ds:
        >>>     print(batch)
    ----
    {'_COL_1': <tf.Tensor: shape=(2,), dtype=float32, numpy=[32.5000, 6.0000]>,
     '_COL_2': <tf.Tensor: shape=(2,), dtype=float32, numpy=[-73.9542, -73.9875]>}
    """
    if not files:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.SNOWML_READ_FAILED,
            original_exception=ValueError("At least one file is needed to create a TF dataset."),
        )

    def generator() -> Generator[Dict[str, npt.NDArray[Any]], None, None]:
        yield from parquet_parser.ParquetParser(list(files), filesystem, batch_size, shuffle, drop_last_batch)

    return tf.data.Dataset.from_generator(generator, output_signature=_derive_signature(files[0], filesystem))


def _arrow_type_to_tensor_spec(field: pa.Field) -> tf.TensorSpec:
    try:
        dtype = tf.dtypes.as_dtype(field.type.to_pandas_dtype())
    except TypeError:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_DATA_TYPE,
            original_exception=TypeError(f"Column {field.name} has unsupportd type {field.type}."),
        )
    # First dimension is batch dimension.
    return tf.TensorSpec(shape=(None,), dtype=dtype)


def _derive_signature(file: str, filesystem: fsspec.AbstractFileSystem) -> Dict[str, tf.TensorSpec]:
    """Derives the signature of the TF dataset from one parquet file."""
    # TODO(zpeng): pq.read_schema does not support `filesystem` until pyarrow>=10.
    # switch to pq.read_schema when we depend on that.
    schema = pq.read_table(file, filesystem=filesystem).schema
    # Signature:
    # The dataset yields dicts. Keys are column names; values are 1-D tensors (
    # the first dimension is batch dimension).
    return {field.name: _arrow_type_to_tensor_spec(field) for field in schema}



import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Set

_PEFT_CONFIG_NAME = "adapter_config.json"


class SupportedLLMType(Enum):
    LLAMA_MODEL_TYPE = "llama"
    OPT_MODEL_TYPE = "opt"

    @classmethod
    def valid_values(cls) -> Set[str]:
        return {member.value for member in cls}


@dataclass(frozen=True)
class LLMOptions:
    """
    This is the option class for LLM.

    Args:
        revision: Revision of HF model. Defaults to None.
        token: The token to use as HTTP bearer authorization for remote files. Defaults to None.
        max_batch_size: Max batch size allowed for single inferenced. Defaults to 1.
    """

    revision: Optional[str] = field(default=None)
    token: Optional[str] = field(default=None)
    max_batch_size: int = field(default=1)
    enable_tp: bool = field(default=False)
    # TODO(halu): Below could be per query call param instead.
    temperature: float = field(default=0.01)
    top_p: float = field(default=1.0)
    max_tokens: int = field(default=100)


class LLM:
    class Mode(Enum):
        LOCAL_LORA = "local_lora"
        REMOTE_PRETRAIN = "remote_pretrain"

    def __init__(
        self,
        model_id_or_path: str,
        *,
        options: Optional[LLMOptions] = None,
    ) -> None:
        """

        Args:
            model_id_or_path: model_id or local dir to PEFT lora weights.
            options: Options for LLM. Defaults to be None.

        Raises:
            ValueError: When unsupported.
        """
        if not options:
            options = LLMOptions()
        hub_kwargs = {
            "revision": options.revision,
            "token": options.token,
        }
        import transformers

        if os.path.isdir(model_id_or_path):
            if not os.path.isfile(os.path.join(model_id_or_path, _PEFT_CONFIG_NAME)):
                raise ValueError("Peft config is not found.")

            import peft

            peft_config = peft.PeftConfig.from_pretrained(model_id_or_path, **hub_kwargs)  # type: ignore[attr-defined]
            if peft_config.peft_type != peft.PeftType.LORA:  # type: ignore[attr-defined]
                raise ValueError("Only LORA is supported.")
            if peft_config.task_type != peft.TaskType.CAUSAL_LM:  # type: ignore[attr-defined]
                raise ValueError("Only CAUSAL_LM is supported.")
            base_model = peft_config.base_model_name_or_path
            base_config = transformers.AutoConfig.from_pretrained(base_model, **hub_kwargs)
            assert (
                base_config.model_type in SupportedLLMType.valid_values()
            ), f"{base_config.model_type} is not supported."
            self.mode = LLM.Mode.LOCAL_LORA
            self.model_type = base_config.model_type
        else:
            # We support pre-train model as well
            model_config = transformers.AutoConfig.from_pretrained(
                model_id_or_path,
                **hub_kwargs,
            )
            assert (
                model_config.model_type in SupportedLLMType.valid_values()
            ), f"{model_config.model_type} is not supported."
            self.mode = LLM.Mode.REMOTE_PRETRAIN
            self.model_type = model_config.model_type

        self.model_id_or_path = model_id_or_path
        self.token = options.token
        self.revision = options.revision
        self.max_batch_size = options.max_batch_size
        self.temperature = options.temperature
        self.top_p = options.top_p
        self.max_tokens = options.max_tokens
        self.enable_tp = options.enable_tp



from types import ModuleType
from typing import Any, Dict, List, Literal, Optional, Union, cast, overload

import pandas as pd

from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import identifier
from snowflake.ml.model import (
    deploy_platforms,
    model_signature,
    type_hints as model_types,
)
from snowflake.ml.model._deploy_client.snowservice import deploy as snowservice_deploy
from snowflake.ml.model._deploy_client.utils import constants as snowservice_constants
from snowflake.ml.model._deploy_client.warehouse import (
    deploy as warehouse_deploy,
    infer_template,
)
from snowflake.ml.model._model_composer import model_composer
from snowflake.ml.model._signatures import snowpark_handler
from snowflake.snowpark import DataFrame as SnowparkDataFrame, Session, functions as F


@overload
def save_model(
    *,
    name: str,
    model: model_types.SupportedNoSignatureRequirementsModelType,
    session: Session,
    stage_path: str,
    metadata: Optional[Dict[str, str]] = None,
    conda_dependencies: Optional[List[str]] = None,
    pip_requirements: Optional[List[str]] = None,
    python_version: Optional[str] = None,
    ext_modules: Optional[List[ModuleType]] = None,
    code_paths: Optional[List[str]] = None,
    options: Optional[model_types.ModelSaveOption] = None,
) -> model_composer.ModelComposer:
    """Save a model that does not require a signature as model to a stage path.

    Args:
        name: Name of the model.
        model: Model object.
        session: Snowpark connection session.
        stage_path: Path to the stage where model will be saved.
        metadata: Model metadata.
        conda_dependencies: List of Conda package specs. Use "[channel::]package [operator version]" syntax to specify
            a dependency. It is a recommended way to specify your dependencies using conda. When channel is not
            specified, defaults channel will be used. When deploying to Snowflake Warehouse, defaults channel would be
            replaced with the Snowflake Anaconda channel.
        pip_requirements: List of PIP package specs. Model will not be able to deploy to the warehouse if there is pip
            requirements.
        python_version: A string of python version where model is run. Used for user override. If specified as None,
            current version would be captured. Defaults to None.
        code_paths: Directory of code to import.
        ext_modules: External modules that user might want to get pickled with model object. Defaults to None.
        options: Model specific kwargs.
    """
    ...


@overload
def save_model(
    *,
    name: str,
    model: model_types.SupportedRequireSignatureModelType,
    session: Session,
    stage_path: str,
    signatures: Dict[str, model_signature.ModelSignature],
    metadata: Optional[Dict[str, str]] = None,
    conda_dependencies: Optional[List[str]] = None,
    pip_requirements: Optional[List[str]] = None,
    python_version: Optional[str] = None,
    ext_modules: Optional[List[ModuleType]] = None,
    code_paths: Optional[List[str]] = None,
    options: Optional[model_types.ModelSaveOption] = None,
) -> model_composer.ModelComposer:
    """Save a model that requires a external signature with user provided signatures as model to a stage path.

    Args:
        name: Name of the model.
        model: Model object.
        session: Snowpark connection session.
        stage_path: Path to the stage where model will be saved.
        signatures: Model data signatures for inputs and output for every target methods.
        metadata: Model metadata.
        conda_dependencies: List of Conda package specs. Use "[channel::]package [operator version]" syntax to specify
            a dependency. It is a recommended way to specify your dependencies using conda. When channel is not
            specified, defaults channel will be used. When deploying to Snowflake Warehouse, defaults channel would be
            replaced with the Snowflake Anaconda channel.
        pip_requirements: List of PIP package specs. Model will not be able to deploy to the warehouse if there is pip
            requirements.
        python_version: A string of python version where model is run. Used for user override. If specified as None,
            current version would be captured. Defaults to None.
        code_paths: Directory of code to import.
        ext_modules: External modules that user might want to get pickled with model object. Defaults to None.
        options: Model specific kwargs.
    """
    ...


@overload
def save_model(
    *,
    name: str,
    model: model_types.SupportedRequireSignatureModelType,
    session: Session,
    stage_path: str,
    sample_input: model_types.SupportedDataType,
    metadata: Optional[Dict[str, str]] = None,
    conda_dependencies: Optional[List[str]] = None,
    pip_requirements: Optional[List[str]] = None,
    python_version: Optional[str] = None,
    ext_modules: Optional[List[ModuleType]] = None,
    code_paths: Optional[List[str]] = None,
    options: Optional[model_types.ModelSaveOption] = None,
) -> model_composer.ModelComposer:
    """Save a model that requires a external signature as model to a stage path with signature inferred from a
      sample_input_data.

    Args:
        name: Name of the model.
        model: Model object.
        session: Snowpark connection session.
        stage_path: Path to the stage where model will be saved.
        sample_input: Sample input data to infer the model signatures from.
        metadata: Model metadata.
        conda_dependencies: List of Conda package specs. Use "[channel::]package [operator version]" syntax to specify
            a dependency. It is a recommended way to specify your dependencies using conda. When channel is not
            specified, defaults channel will be used. When deploying to Snowflake Warehouse, defaults channel would be
            replaced with the Snowflake Anaconda channel.
        pip_requirements: List of PIP package specs. Model will not be able to deploy to the warehouse if there is pip
            requirements.
        python_version: A string of python version where model is run. Used for user override. If specified as None,
            current version would be captured. Defaults to None.
        code_paths: Directory of code to import.
        ext_modules: External modules that user might want to get pickled with model object. Defaults to None.
        options: Model specific kwargs.
    """
    ...


def save_model(
    *,
    name: str,
    model: model_types.SupportedModelType,
    session: Session,
    stage_path: str,
    signatures: Optional[Dict[str, model_signature.ModelSignature]] = None,
    sample_input: Optional[model_types.SupportedDataType] = None,
    metadata: Optional[Dict[str, str]] = None,
    conda_dependencies: Optional[List[str]] = None,
    pip_requirements: Optional[List[str]] = None,
    python_version: Optional[str] = None,
    ext_modules: Optional[List[ModuleType]] = None,
    code_paths: Optional[List[str]] = None,
    options: Optional[model_types.ModelSaveOption] = None,
) -> model_composer.ModelComposer:
    """Save the model.

    Args:
        name: Name of the model.
        model: Model object.
        session: Snowpark connection session.
        stage_path: Path to the stage where model will be saved.
        signatures: Model data signatures for inputs and output for every target methods. If it is None, sample_input
            would be used to infer the signatures if it is a local (non-SnowML modeling model).
            If not None, sample_input should not be specified. Defaults to None.
        sample_input: Sample input data to infer the model signatures from. If it is None, signatures must be specified
            if it is a local (non-SnowML modeling model). If not None, signatures should not be specified.
            Defaults to None.
        metadata: Model metadata.
        conda_dependencies: List of Conda package specs. Use "[channel::]package [operator version]" syntax to specify
            a dependency. It is a recommended way to specify your dependencies using conda. When channel is not
            specified, defaults channel will be used. When deploying to Snowflake Warehouse, defaults channel would be
            replaced with the Snowflake Anaconda channel.
        pip_requirements: List of PIP package specs. Model will not be able to deploy to the warehouse if there is pip
            requirements.
        python_version: A string of python version where model is run. Used for user override. If specified as None,
            current version would be captured. Defaults to None.
        code_paths: Directory of code to import.
        ext_modules: External modules that user might want to get pickled with model object. Defaults to None.
        options: Model specific kwargs.

    Returns:
        Model
    """
    m = model_composer.ModelComposer(session=session, stage_path=stage_path)
    m.save(
        name=name,
        model=model,
        signatures=signatures,
        sample_input=sample_input,
        metadata=metadata,
        conda_dependencies=conda_dependencies,
        pip_requirements=pip_requirements,
        python_version=python_version,
        ext_modules=ext_modules,
        code_paths=code_paths,
        options=options,
    )
    return m


@overload
def load_model(*, session: Session, stage_path: str) -> model_composer.ModelComposer:
    """Load the model into memory from a zip file in the stage.

    Args:
        session: Snowflake connection session.
        stage_path: Path to the stage where model will be loaded from.
    """
    ...


@overload
def load_model(*, session: Session, stage_path: str, meta_only: Literal[False]) -> model_composer.ModelComposer:
    """Load the model into memory from a zip file in the stage.

    Args:
        session: Snowflake connection session.
        stage_path: Path to the stage where model will be loaded from.
        meta_only: Flag to indicate that if only load metadata.
    """
    ...


@overload
def load_model(*, session: Session, stage_path: str, meta_only: Literal[True]) -> model_composer.ModelComposer:
    """Load the model into memory from a zip file in the stage with metadata only.

    Args:
        session: Snowflake connection session.
        stage_path: Path to the stage where model will be loaded from.
        meta_only: Flag to indicate that if only load metadata.
    """
    ...


def load_model(
    *,
    session: Session,
    stage_path: str,
    meta_only: bool = False,
) -> model_composer.ModelComposer:
    """Load the model into memory from directory or a zip file in the stage.

    Args:
        session: Snowflake connection session. Must be specified when specifying model_stage_file_path.
            Exclusive with model_dir_path.
        stage_path: Path to the stage where model will be loaded from.
        meta_only: Flag to indicate that if only load metadata.

    Returns:
        Loaded model.
    """
    m = model_composer.ModelComposer(session=session, stage_path=stage_path)
    m.load(meta_only=meta_only)
    return m


@overload
def deploy(
    session: Session,
    *,
    name: str,
    platform: deploy_platforms.TargetPlatform,
    target_method: Optional[str],
    stage_path: str,
    options: Optional[model_types.DeployOptions],
) -> Optional[model_types.Deployment]:
    """Create a deployment from a model in a zip file in a stage and deploy it to remote platform.

    Args:
        session: Snowpark Connection Session.
        name: Name of the deployment for the model.
        platform: Target platform to deploy the model.
        target_method: The name of the target method to be deployed. Can be omitted if there is only 1 target method in
            the model.
        stage_path: Path to the stage where model will be deployed.
        options: Additional options when deploying the model.
            Each target platform will have their own specifications of options.
    """
    ...


@overload
def deploy(
    session: Session,
    *,
    model_id: str,
    name: str,
    platform: deploy_platforms.TargetPlatform,
    target_method: Optional[str],
    stage_path: str,
    deployment_stage_path: str,
    options: Optional[model_types.DeployOptions],
) -> Optional[model_types.Deployment]:
    """Create a deployment from a model in a local directory and deploy it to remote platform.

    Args:
        session: Snowpark Connection Session.
        model_id: Internal model ID string.
        name: Name of the deployment for the model.
        platform: Target platform to deploy the model.
        target_method: The name of the target method to be deployed. Can be omitted if there is only 1 target method in
            the model.
        stage_path: Path to the stage where model will be deployed.
        deployment_stage_path: Path to stage containing snowpark container service deployment artifacts.
        options: Additional options when deploying the model.
            Each target platform will have their own specifications of options.
    """
    ...


def deploy(
    session: Session,
    *,
    name: str,
    platform: deploy_platforms.TargetPlatform,
    stage_path: str,
    target_method: Optional[str] = None,
    deployment_stage_path: Optional[str] = None,
    model_id: Optional[str] = None,
    options: Optional[model_types.DeployOptions],
) -> Optional[model_types.Deployment]:
    """Create a deployment from a model and deploy it to remote platform.

    Args:
        session: Snowpark Connection Session.
        model_id: Internal model ID string.
        name: Name of the deployment for the model.
        platform: Target platform to deploy the model.
        target_method: The name of the target method to be deployed. Can be omitted if there is only 1 target method in
            the model.
        stage_path: Path to the stage where model will be deployed.
        deployment_stage_path: Path to stage containing deployment artifacts.
        options: Additional options when deploying the model.
            Each target platform will have their own specifications of options.

    Raises:
        SnowflakeMLException: Raised when target platform is unsupported.
        SnowflakeMLException: Raised when target method does not exist in model.

    Returns:
        The deployment information.
    """

    info = None

    if not options:
        options = {}

    m = load_model(session=session, stage_path=stage_path, meta_only=True)
    assert m.packager.meta

    if target_method is None:
        if len(m.packager.meta.signatures.keys()) == 1:
            target_method = list(m.packager.meta.signatures.keys())[0]
        else:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    "Only when the model has 1 target methods can target_method be omitted when deploying."
                ),
            )

    details: model_types.DeployDetails = {}
    if platform == deploy_platforms.TargetPlatform.WAREHOUSE:
        warehouse_deploy._deploy_to_warehouse(
            session=session,
            model_stage_file_path=m.model_stage_path,
            model_meta=m.packager.meta,
            udf_name=name,
            target_method=target_method,
            **options,
        )

    elif platform == deploy_platforms.TargetPlatform.SNOWPARK_CONTAINER_SERVICES:
        options = cast(model_types.SnowparkContainerServiceDeployOptions, options)
        assert model_id, "Require 'model_id' for Snowpark container service deployment"
        assert m.model_stage_path, "Require 'model_stage_file_path' for Snowpark container service deployment"
        assert deployment_stage_path, "Require 'deployment_stage_path' for Snowpark container service deployment"
        if snowservice_constants.COMPUTE_POOL not in options:
            raise snowml_exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=ValueError(
                    "Missing 'compute_pool' in options field for Snowpark container service deployment"
                ),
            )

        details = snowservice_deploy._deploy(
            session=session,
            model_id=model_id,
            model_meta=m.packager.meta,
            service_func_name=name,
            model_zip_stage_path=m.model_stage_path,
            deployment_stage_path=deployment_stage_path,
            target_method=target_method,
            **options,
        )

    else:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_TYPE,
            original_exception=ValueError(f"Unsupported target Platform: {platform}"),
        )
    signature = m.packager.meta.signatures.get(target_method, None)
    if not signature:
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INVALID_ARGUMENT,
            original_exception=ValueError(f"Target method {target_method} does not exist in model."),
        )
    info = model_types.Deployment(
        name=name, platform=platform, target_method=target_method, signature=signature, options=options, details=details
    )
    return info


@overload
def predict(
    session: Session,
    *,
    deployment: model_types.Deployment,
    X: model_types.SupportedLocalDataType,
    statement_params: Optional[Dict[str, Any]] = None,
) -> pd.DataFrame:
    """Execute batch inference of a model remotely on local data. Can be any supported data type. Return a local
        Pandas Dataframe.

    Args:
        session: Snowpark Connection Session.
        deployment: The deployment info to use for predict.
        X: The input data.
        statement_params: Statement Parameters for telemetry.
    """
    ...


@overload
def predict(
    session: Session,
    *,
    deployment: model_types.Deployment,
    X: SnowparkDataFrame,
    statement_params: Optional[Dict[str, Any]] = None,
) -> SnowparkDataFrame:
    """Execute batch inference of a model remotely on a Snowpark DataFrame. Return a Snowpark DataFrame.

    Args:
        session: Snowpark Connection Session.
        deployment: The deployment info to use for predict.
        X: The input Snowpark dataframe.
        statement_params: Statement Parameters for telemetry.
    """
    ...


def predict(
    session: Session,
    *,
    deployment: model_types.Deployment,
    X: Union[model_types.SupportedDataType, SnowparkDataFrame],
    statement_params: Optional[Dict[str, Any]] = None,
) -> Union[pd.DataFrame, SnowparkDataFrame]:
    """Execute batch inference of a model remotely.

    Args:
        session: Snowpark Connection Session.
        deployment: The deployment info to use for predict.
        X: The input dataframe.
        statement_params: Statement Parameters for telemetry.

    Returns:
        The output dataframe.
    """

    # Get options
    INTERMEDIATE_OBJ_NAME = "tmp_result"
    sig = deployment["signature"]

    # Validate and prepare input
    if not isinstance(X, SnowparkDataFrame):
        keep_order = True
        output_with_input_features = False
        df = model_signature._convert_and_validate_local_data(X, sig.inputs)
        s_df = snowpark_handler.SnowparkDataFrameHandler.convert_from_df(session, df, keep_order=keep_order)
    else:
        keep_order = False
        output_with_input_features = True
        model_signature._validate_snowpark_data(X, sig.inputs)
        s_df = X

    if statement_params:
        if s_df._statement_params is not None:
            s_df._statement_params.update(statement_params)
        else:
            s_df._statement_params = statement_params  # type: ignore[assignment]

    # Infer and get intermediate result
    input_cols = []
    for col_name in s_df.columns:
        literal_col_name = identifier.get_unescaped_names(col_name)
        input_cols.extend(
            [
                F.lit(literal_col_name),
                F.col(col_name),
            ]
        )

    # TODO[shchen]: SNOW-870032, For SnowService, external function name cannot be double quoted, else it results in
    # external function no found.
    udf_name = deployment["name"]
    output_obj = F.call_udf(udf_name, F.object_construct(*input_cols))

    if output_with_input_features:
        df_res = s_df.with_column(INTERMEDIATE_OBJ_NAME, output_obj)
    else:
        df_res = s_df.select(output_obj.alias(INTERMEDIATE_OBJ_NAME))

    if keep_order:
        df_res = df_res.order_by(
            F.col(INTERMEDIATE_OBJ_NAME)[infer_template._KEEP_ORDER_COL_NAME],
            ascending=True,
        )

    # Prepare the output
    output_cols = []
    for output_feature in sig.outputs:
        output_cols.append(F.col(INTERMEDIATE_OBJ_NAME)[output_feature.name].astype(output_feature.as_snowpark_type()))

    df_res = df_res.with_columns(
        [identifier.get_inferred_name(output_feature.name) for output_feature in sig.outputs],
        output_cols,
    ).drop(INTERMEDIATE_OBJ_NAME)

    # Get final result
    if not isinstance(X, SnowparkDataFrame):
        return snowpark_handler.SnowparkDataFrameHandler.convert_to_df(df_res, features=sig.outputs)
    else:
        return df_res



import math
from contextlib import contextmanager
from timeit import default_timer
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional

import snowflake.snowpark.functions as F
from snowflake import snowpark


@contextmanager
def timer() -> Generator[Callable[[], float], None, None]:
    start: float = default_timer()

    def elapser() -> float:
        return default_timer() - start

    yield lambda: elapser()


def _flatten(L: Iterable[List[Any]]) -> List[Any]:
    return [val for sublist in L for val in sublist]


def map_dataframe_by_column(
    df: snowpark.DataFrame,
    cols: List[str],
    map_func: Callable[[snowpark.DataFrame, List[str]], snowpark.DataFrame],
    partition_size: int,
    statement_params: Optional[Dict[str, Any]] = None,
) -> List[List[Any]]:
    """Applies the `map_func` to the input DataFrame by parallelizing it over subsets of the column.

    Because the return results are materialized as Python lists *in memory*, this method should
    not be used on operations that are expected to return many rows.

    The `map_func` must satisfy the property that for an input `df` with columns C, then for any
    partition [*c1, *c2, ..., *cp] on C, then `map_func(df,C) == [*map_func(df,c1), ..., *map_func(df,cp)]`,
    where * is the list unpacking operator. This means that `map_func(df, col_subset)` should
    return r rows and c*|col_subset| columns, for constants r and c.

    Args:
        df: Input dataset to operate on.
        cols: List of column names to compute on. Must index into `df`.
        map_func: The map function applied on each partition of the DataFrame.
        partition_size: The number of columns to include in each partition. Must be a positive integer.
        statement_params: Statement parameters for query telemetry.

    Returns:
        A Python list representation of the output of the query.

    Raises:
        Exception: If the pre-conditions above are not met.
    """
    partition_id_col = "_PARTITION_ID"
    n_output_cols = 0
    unioned_df: Optional[snowpark.DataFrame] = None
    last_partition_df = None

    if partition_size < 1:
        raise Exception(f"Partition size must be a positive integer, but got {partition_size}.")

    try:
        n_partitions = math.ceil(len(df[cols].columns) / partition_size)
    except Exception:
        raise Exception(f"Provided column names {cols} does not index into the dataset.")

    # This should never happen
    if n_partitions == 0:
        return [[]]

    # Create one DataFrame for the first n-1 partitions, and one for the last partition.
    for partition_id in range(n_partitions):
        cols_subset = cols[(partition_id * partition_size) : ((partition_id + 1) * partition_size)]
        mapped_df = map_func(df, cols_subset)
        if partition_id == 0:
            n_output_cols = len(mapped_df.columns)

        if partition_id == n_partitions - 1:
            last_partition_df = mapped_df
        else:
            if n_output_cols != len(mapped_df.columns):
                raise Exception("All partitions must contain the same number of columns.")
            mapped_df = mapped_df.with_column(partition_id_col, F.lit(partition_id))
            unioned_df = mapped_df if unioned_df is None else unioned_df.union_all(mapped_df)

    # Store results in a list of size |n_partitions| x |n_rows| x |n_output_cols|
    all_results: List[List[List[Any]]] = [[] for _ in range(n_partitions - 1)]

    # Collect the results of the first n-1 partitions, removing the partition_id column
    unioned_result = unioned_df.collect(statement_params=statement_params) if unioned_df is not None else []
    for row in unioned_result:
        row_dict = row.as_dict()
        partition_id = row_dict.pop(partition_id_col, None)
        if partition_id is None:
            raise Exception(f"Found unknown partition id {partition_id}.")
        all_results[partition_id].append(list(row_dict.values()))

    # Collect the results of the last partition
    last_partition_result = (
        [[]]
        if last_partition_df is None
        else [list(row) for row in last_partition_df.collect(statement_params=statement_params)]
    )
    all_results.append(last_partition_result)

    row_counts = {len(res) for res in all_results}
    if len(row_counts) > 1:
        raise Exception(
            f"All partitions must return the same number of rows, but found multiple row counts: {row_counts}."
        )

    return [_flatten(row) for row in list(zip(*all_results))]



import asyncio
import http
import logging
import os
import sys
import tempfile
import threading
import time
import traceback
import zipfile
from enum import Enum
from typing import Dict, List, Optional, cast

import pandas as pd
from gunicorn import arbiter
from starlette import applications, concurrency, requests, responses, routing


class _ModelLoadingState(Enum):
    """
    Enum class to represent various model loading state.
    """

    LOADING = "loading"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class CustomThread(threading.Thread):
    """
    Custom Thread implementation that overrides Thread.run.

    This is necessary because the default Thread implementation suppresses exceptions in child threads. The standard
    behavior involves the Thread class catching exceptions and throwing a SystemExit exception, which requires
    Thread.join to terminate the process. To address this, we overwrite Thread.run and use os._exit instead.

    We throw specific error code "Arbiter.APP_LOAD_ERROR" such that Gunicorn Arbiter master process will be killed,
    which then trigger the container to be marked as failed. This ensures the container becomes ready when all workers
    loaded the model successfully.
    """

    def run(self) -> None:
        try:
            super().run()
            # Keep the daemon thread alive to avoid destroy
            # state attached to thread when loading the model.
            while True:
                time.sleep(60)
        except Exception:
            logger.error(traceback.format_exc())
            os._exit(arbiter.Arbiter.APP_LOAD_ERROR)


logger = logging.getLogger(__name__)
_LOADED_MODEL = None
_LOADED_META = None
_MODEL_CODE_DIR = "code"
_MODEL_LOADING_STATE = _ModelLoadingState.LOADING
_MODEL_LOADING_EVENT = threading.Event()
_CONCURRENT_REQUESTS_MAX: Optional[int] = None
_CONCURRENT_COUNTER = 0
_CONCURRENT_COUNTER_LOCK = asyncio.Lock()
TARGET_METHOD = None


def _run_setup() -> None:
    """Set up logging and load model into memory."""
    # Align the application logger's handler with Gunicorn's to capture logs from all processes.
    gunicorn_logger = logging.getLogger("gunicorn.error")
    logger.handlers = gunicorn_logger.handlers
    logger.setLevel(gunicorn_logger.level)

    logger.info(f"ENV: {os.environ}")

    global _LOADED_MODEL
    global _LOADED_META
    global _MODEL_LOADING_STATE
    global _MODEL_LOADING_EVENT
    global _CONCURRENT_REQUESTS_MAX
    global TARGET_METHOD

    try:
        model_zip_stage_path = os.getenv("MODEL_ZIP_STAGE_PATH")
        assert model_zip_stage_path, "Missing environment variable MODEL_ZIP_STAGE_PATH"

        TARGET_METHOD = os.getenv("TARGET_METHOD")

        _concurrent_requests_max_env = os.getenv("_CONCURRENT_REQUESTS_MAX", None)

        _CONCURRENT_REQUESTS_MAX = int(_concurrent_requests_max_env) if _concurrent_requests_max_env else None

        with tempfile.TemporaryDirectory() as tmp_dir:
            if zipfile.is_zipfile(model_zip_stage_path):
                extracted_dir = os.path.join(tmp_dir, "extracted_model_dir")
                logger.info(f"Extracting model zip from {model_zip_stage_path} to {extracted_dir}")
                with zipfile.ZipFile(model_zip_stage_path, "r") as model_zip:
                    if len(model_zip.namelist()) > 1:
                        model_zip.extractall(extracted_dir)
            else:
                raise RuntimeError(f"No model zip found at stage path: {model_zip_stage_path}")
            logger.info(f"Loading model from {extracted_dir} into memory")

            sys.path.insert(0, os.path.join(extracted_dir, _MODEL_CODE_DIR))
            from snowflake.ml.model import type_hints as model_types

            # TODO (Server-side Model Rollout):
            # Keep try block only
            try:
                from snowflake.ml.model._packager import model_packager

                pk = model_packager.ModelPackager(extracted_dir)
                pk.load(
                    as_custom_model=True,
                    meta_only=False,
                    options=model_types.ModelLoadOption(
                        {"use_gpu": cast(bool, os.environ.get("SNOWML_USE_GPU", False))}
                    ),
                )
                _LOADED_MODEL = pk.model
                _LOADED_META = pk.meta
            except ImportError as e:
                if e.name and not e.name.startswith("snowflake.ml"):
                    raise e
                # Legacy model support
                from snowflake.ml.model import (  # type: ignore[attr-defined]
                    _model as model_api,
                )

                if hasattr(model_api, "_load_model_for_deploy"):
                    _LOADED_MODEL, _LOADED_META = model_api._load_model_for_deploy(extracted_dir)
                else:
                    _LOADED_MODEL, meta_LOADED_META = model_api._load(
                        local_dir_path=extracted_dir,
                        as_custom_model=True,
                        options=model_types.ModelLoadOption(
                            {"use_gpu": cast(bool, os.environ.get("SNOWML_USE_GPU", False))}
                        ),
                    )
            _MODEL_LOADING_STATE = _ModelLoadingState.SUCCEEDED
            logger.info("Successfully loaded model into memory")
            _MODEL_LOADING_EVENT.set()
    except Exception as e:
        _MODEL_LOADING_STATE = _ModelLoadingState.FAILED
        raise e


async def ready(request: requests.Request) -> responses.JSONResponse:
    """Check if the application is ready to serve requests.

    This endpoint is used to determine the readiness of the application to handle incoming requests. It returns an HTTP
    200 status code only when the model has been successfully loaded into memory. If the model has not yet been loaded,
    it responds with an HTTP 503 status code, which signals to the readiness probe to continue probing until the
    application becomes ready or until the client's timeout is reached.

    Args:
        request:
            The HTTP request object.

    Returns:
        A JSON response with status information:
        - HTTP 200 status code and {"status": "ready"} when the model is loaded and the application is ready.
        - HTTP 503 status code and {"status": "not ready"} when the model is not yet loaded.

    """
    if _MODEL_LOADING_STATE == _ModelLoadingState.SUCCEEDED:
        return responses.JSONResponse({"status": "ready"})
    return responses.JSONResponse({"status": "not ready"}, status_code=http.HTTPStatus.SERVICE_UNAVAILABLE)


def _do_predict(input_json: Dict[str, List[List[object]]]) -> responses.JSONResponse:
    from snowflake.ml.model.model_signature import FeatureSpec

    assert _LOADED_MODEL, "model is not loaded"
    assert _LOADED_META, "model metadata is not loaded"
    assert TARGET_METHOD, "Missing environment variable TARGET_METHOD"

    try:
        features = cast(List[FeatureSpec], _LOADED_META.signatures[TARGET_METHOD].inputs)
        dtype_map = {feature.name: feature.as_dtype() for feature in features}
        input_cols = [spec.name for spec in features]
        output_cols = [spec.name for spec in _LOADED_META.signatures[TARGET_METHOD].outputs]
        assert "data" in input_json, "missing data field in the request input"
        # The expression x[1:] is used to exclude the index of the data row.
        input_data = [x[1] for x in input_json["data"]]
        df = pd.json_normalize(input_data).astype(dtype=dtype_map)
        x = df[input_cols]
        assert len(input_data) != 0 and not all(not row for row in input_data), "empty data"
    except Exception as e:
        error_message = f"Input data malformed: {str(e)}\n{traceback.format_exc()}"
        logger.error(f"Failed request with error: {error_message}")
        return responses.JSONResponse({"error": error_message}, status_code=http.HTTPStatus.BAD_REQUEST)

    try:
        predictions_df = getattr(_LOADED_MODEL, TARGET_METHOD)(x)
        predictions_df.columns = output_cols
        # Use _ID to keep the order of prediction result and associated features.
        _KEEP_ORDER_COL_NAME = "_ID"
        if _KEEP_ORDER_COL_NAME in df.columns:
            predictions_df[_KEEP_ORDER_COL_NAME] = df[_KEEP_ORDER_COL_NAME]
        response = {"data": [[i, row] for i, row in enumerate(predictions_df.to_dict(orient="records"))]}
        return responses.JSONResponse(response)
    except Exception as e:
        error_message = f"Prediction failed: {str(e)}\n{traceback.format_exc()}"
        logger.error(f"Failed request with error: {error_message}")
        return responses.JSONResponse({"error": error_message}, status_code=http.HTTPStatus.BAD_REQUEST)


async def predict(request: requests.Request) -> responses.JSONResponse:
    """Endpoint to make predictions based on input data.

    Args:
        request: The input data is expected to be in the following JSON format:
            {
                "data": [
                    [0, {'_ID': 0, 'input_feature_0': 0.0, 'input_feature_1': 1.0}],
                    [1, {'_ID': 1, 'input_feature_0': 2.0, 'input_feature_1': 3.0}],
            }
            Each row is represented as a list, where the first element denotes the index of the row.

    Returns:
        Two possible responses:
        For success, return a JSON response
            {
                "data": [
                    [0, {'_ID': 0, 'output': 1}],
                    [1, {'_ID': 1, 'output': 2}]
                ]
            },
            The first element of each resulting list denotes the index of the row, and the rest of the elements
            represent the prediction results for that row.
        For an error, return {"error": error_message, "status_code": http_response_status_code}.
    """
    _MODEL_LOADING_EVENT.wait()  # Ensure model is indeed loaded into memory

    global _CONCURRENT_COUNTER
    global _CONCURRENT_COUNTER_LOCK

    input_json = await request.json()

    if _CONCURRENT_REQUESTS_MAX:
        async with _CONCURRENT_COUNTER_LOCK:
            if _CONCURRENT_COUNTER >= int(_CONCURRENT_REQUESTS_MAX):
                return responses.JSONResponse(
                    {"error": "Too many requests"}, status_code=http.HTTPStatus.TOO_MANY_REQUESTS
                )

    async with _CONCURRENT_COUNTER_LOCK:
        _CONCURRENT_COUNTER += 1

    resp = await concurrency.run_in_threadpool(_do_predict, input_json)

    async with _CONCURRENT_COUNTER_LOCK:
        _CONCURRENT_COUNTER -= 1

    return resp


def run_app() -> applications.Starlette:
    # TODO[shchen]: SNOW-893654. Before SnowService supports Startup probe, or extends support for Readiness probe
    # with configurable failureThreshold, we will have to load the model in a separate thread in order to prevent
    # gunicorn worker timeout.
    model_loading_worker = CustomThread(target=_run_setup, daemon=True)
    model_loading_worker.start()

    routes = [
        routing.Route("/health", endpoint=ready, methods=["GET"]),
        routing.Route("/predict", endpoint=predict, methods=["POST"]),
    ]
    return applications.Starlette(routes=routes)


app = run_app()



#!/usr/bin/env python3
import enum
import functools
import inspect
import operator
import types
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import ParamSpec

from snowflake import connector
from snowflake.connector import telemetry as connector_telemetry, time_util
from snowflake.ml._internal import env
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.snowpark import dataframe, exceptions as snowpark_exceptions, session
from snowflake.snowpark._internal import utils

_log_counter = 0
_FLUSH_SIZE = 10

_Args = ParamSpec("_Args")
_ReturnValue = TypeVar("_ReturnValue")


@enum.unique
class TelemetryField(enum.Enum):
    # constants
    NAME = "name"
    # types of telemetry
    TYPE_FUNCTION_USAGE = "function_usage"
    # message keys for telemetry
    KEY_PROJECT = "project"
    KEY_SUBPROJECT = "subproject"
    KEY_FUNC_NAME = "func_name"
    KEY_FUNC_PARAMS = "func_params"
    KEY_ERROR_INFO = "error_info"
    KEY_ERROR_CODE = "error_code"
    KEY_VERSION = "version"
    KEY_PYTHON_VERSION = "python_version"
    KEY_OS = "operating_system"
    KEY_DATA = "data"
    KEY_CATEGORY = "category"
    KEY_API_CALLS = "api_calls"
    KEY_SFQIDS = "sfqids"
    KEY_CUSTOM_TAGS = "custom_tags"
    # function categories
    FUNC_CAT_USAGE = "usage"


def get_statement_params(
    project: str, subproject: Optional[str] = None, class_name: Optional[str] = None
) -> Dict[str, Any]:
    """
    Get telemetry statement parameters.

    Args:
        project: Project name.
        subproject: Subproject name.
        class_name: Caller's module name.

    Returns:
        dictionary of parameters to log to event table.
    """
    frame = inspect.currentframe()
    return get_function_usage_statement_params(
        project=project,
        subproject=subproject,
        function_name=get_statement_params_full_func_name(frame.f_back if frame else None, class_name),
    )


# TODO: we can merge this with get_statement_params after code clean up
def get_statement_params_full_func_name(frame: Optional[types.FrameType], class_name: Optional[str] = None) -> str:
    """
    Get the class-level or module-level full function name to be logged in statement parameters.
    The full function name is in the form of "module_name.class_name.function_name" (class-level)
    or "module_name.function_name" (module-level) when `class_name` is None.

    Args:
        frame: Frame object for the caller’s stack frame.
        class_name: Class name.

    Returns:
        Full function name.

    Examples:

        >>> import inspect

        >>> func_name = get_statement_params_full_func_name(inspect.currentframe(), "ClassName")
        >>> statement_params = get_function_usage_statement_params(function_name=func_name, ...)
    """
    module = inspect.getmodule(frame)
    module_name = module.__name__ if module else None
    function_name = frame.f_code.co_name if frame else None
    func_name = ".".join([name for name in [module_name, class_name, function_name] if name])
    return func_name


def get_function_usage_statement_params(
    project: Optional[str] = None,
    subproject: Optional[str] = None,
    *,
    function_category: str = TelemetryField.FUNC_CAT_USAGE.value,
    function_name: Optional[str] = None,
    function_parameters: Optional[Dict[str, Any]] = None,
    api_calls: Optional[
        List[
            Union[
                Dict[str, Union[Callable[..., Any], str]],
                Union[Callable[..., Any], str],
            ]
        ]
    ] = None,
    custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
) -> Dict[str, Any]:
    """
    Get function usage statement parameters.

    Args:
        project: Project.
        subproject: Subproject.
        function_category: Function category.
        function_name: Function name.
        function_parameters: Function parameters.
        api_calls: API calls in the function.
        custom_tags: Custom tags.

    Returns:
        Statement parameters.

    Examples:

        >>> statement_params = get_function_usage_statement_params(...)
        >>> df.collect(statement_params=statement_params)
    """
    telemetry_type = f"{env.SOURCE.lower()}_{TelemetryField.TYPE_FUNCTION_USAGE.value}"
    statement_params: Dict[str, Any] = {
        connector_telemetry.TelemetryField.KEY_SOURCE.value: env.SOURCE,
        TelemetryField.KEY_PROJECT.value: project,
        TelemetryField.KEY_SUBPROJECT.value: subproject,
        TelemetryField.KEY_OS.value: env.OS,
        TelemetryField.KEY_VERSION.value: env.VERSION,
        TelemetryField.KEY_PYTHON_VERSION.value: env.PYTHON_VERSION,
        connector_telemetry.TelemetryField.KEY_TYPE.value: telemetry_type,
        TelemetryField.KEY_CATEGORY.value: function_category,
    }

    if function_name:
        statement_params[TelemetryField.KEY_FUNC_NAME.value] = function_name
    if function_parameters:
        statement_params[TelemetryField.KEY_FUNC_PARAMS.value] = function_parameters
    if api_calls:
        statement_params[TelemetryField.KEY_API_CALLS.value] = []
        for api_call in api_calls:
            if isinstance(api_call, Dict):
                telemetry_api_call = api_call.copy()
                # convert Callable to str
                for field, api in api_call.items():
                    if callable(api):
                        telemetry_api_call[field] = _get_full_func_name(api)
                statement_params[TelemetryField.KEY_API_CALLS.value].append(telemetry_api_call)
            elif callable(api_call):
                func_name = _get_full_func_name(api_call)
                statement_params[TelemetryField.KEY_API_CALLS.value].append({TelemetryField.NAME.value: func_name})
            else:
                statement_params[TelemetryField.KEY_API_CALLS.value].append({TelemetryField.NAME.value: api_call})
    if custom_tags:
        statement_params[TelemetryField.KEY_CUSTOM_TAGS.value] = custom_tags
    return statement_params


# TODO: fix the type hints here. It should use TypeVars.
def suppress_exceptions(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator used in the telemetry client to suppress exceptions raised by client functions.
    We should continue the business logic when telemetry fails.

    Args:
        func: Telemetry client function.

    Returns:
        Decorator that suppresses exceptions raised by the decorated function.
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        try:
            func(*args, **kwargs)
        except Exception:
            pass

    return wrapper


def send_api_usage_telemetry(
    project: str,
    subproject: Optional[str] = None,
    *,
    func_params_to_log: Optional[Iterable[str]] = None,
    conn_attr_name: Optional[str] = None,
    api_calls_extractor: Optional[
        Callable[
            ...,
            List[
                Union[
                    Dict[str, Union[Callable[..., Any], str]],
                    Union[Callable[..., Any], str],
                ]
            ],
        ]
    ] = None,
    sfqids_extractor: Optional[Callable[..., List[str]]] = None,
    custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
) -> Callable[[Callable[_Args, _ReturnValue]], Callable[_Args, _ReturnValue]]:
    """
    Decorator that sends API usage telemetry.

    Args:
        project: Project.
        subproject: Subproject.
        func_params_to_log: Function parameters to log.
        conn_attr_name: Name of the SnowflakeConnection attribute in `self`.
        api_calls_extractor: Extract API calls from `self`.
        sfqids_extractor: Extract sfqids from `self`.
        custom_tags: Custom tags.

    Returns:
        Decorator that sends function usage telemetry for any call to the decorated function.

    Raises:
        TypeError: If `conn_attr_name` is provided but the conn attribute is not of type SnowflakeConnection.

    # noqa: DAR402
    """

    def decorator(func: Callable[_Args, _ReturnValue]) -> Callable[_Args, _ReturnValue]:
        @functools.wraps(func)
        def wrap(*args: Any, **kwargs: Any) -> _ReturnValue:
            params = _get_func_params(func, func_params_to_log, args, kwargs) if func_params_to_log else None

            # prioritize `conn_attr_name` over the active session
            if conn_attr_name:
                # raise AttributeError if conn attribute does not exist in `self`
                conn = operator.attrgetter(conn_attr_name)(args[0])
                if not isinstance(conn, connector.SnowflakeConnection):
                    raise TypeError(f"Expected a conn object of type SnowflakeConnection, but got {type(conn)}")
            # get an active session
            else:
                try:
                    active_session = next(iter(session._get_active_sessions()))
                # server no default session
                except snowpark_exceptions.SnowparkSessionException:
                    try:
                        return func(*args, **kwargs)
                    except Exception as e:
                        if isinstance(e, snowml_exceptions.SnowflakeMLException):
                            e = e.original_exception
                        # suppress SnowparkSessionException from telemetry in the stack trace
                        raise e from None

                conn = active_session._conn._conn
                if (not active_session.telemetry_enabled) or (conn is None):
                    try:
                        return func(*args, **kwargs)
                    except snowml_exceptions.SnowflakeMLException as e:
                        raise e.original_exception from e

            api_calls: List[Dict[str, Any]] = []
            if api_calls_extractor:
                extracted_api_calls = api_calls_extractor(args[0])
                for api_call in extracted_api_calls:
                    if isinstance(api_call, str):
                        api_calls.append({TelemetryField.NAME.value: api_call})
                    elif callable(api_call):
                        api_calls.append({TelemetryField.NAME.value: _get_full_func_name(api_call)})
                    else:
                        api_calls.append(api_call)
            api_calls.append({TelemetryField.NAME.value: _get_full_func_name(func)})

            sfqids = None
            if sfqids_extractor:
                sfqids = sfqids_extractor(args[0])

            # TODO(hayu): [SNOW-750287] Optimize telemetry client to a singleton.
            telemetry = _SourceTelemetryClient(conn=conn, project=project, subproject=subproject)
            telemetry_args = dict(
                func_name=_get_full_func_name(func),
                function_category=TelemetryField.FUNC_CAT_USAGE.value,
                func_params=params,
                api_calls=api_calls,
                sfqids=sfqids,
                custom_tags=custom_tags,
            )
            try:
                res = func(*args, **kwargs)
            except Exception as e:
                if not isinstance(e, snowml_exceptions.SnowflakeMLException):
                    # already handled via a nested decorated function
                    if hasattr(e, "_snowflake_ml_handled") and e._snowflake_ml_handled:
                        raise e
                    if isinstance(e, snowpark_exceptions.SnowparkClientException):
                        e = snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.INTERNAL_SNOWPARK_ERROR, original_exception=e
                        )
                    else:
                        e = snowml_exceptions.SnowflakeMLException(
                            error_code=error_codes.UNDEFINED, original_exception=e
                        )
                telemetry_args["error"] = repr(e)
                telemetry_args["error_code"] = e.error_code
                e.original_exception._snowflake_ml_handled = True  # type: ignore[attr-defined]
                if e.suppress_source_trace:
                    raise e.original_exception from None
                else:
                    raise e.original_exception from e
            else:
                return res
            finally:
                telemetry.send_function_usage_telemetry(**telemetry_args)
                global _log_counter
                _log_counter += 1
                if _log_counter >= _FLUSH_SIZE or "error" in telemetry_args:
                    telemetry.send_batch()
                    _log_counter = 0

        return cast(Callable[_Args, _ReturnValue], wrap)

    return decorator


def add_stmt_params_to_df(
    project: str,
    subproject: Optional[str] = None,
    *,
    function_category: str = TelemetryField.FUNC_CAT_USAGE.value,
    func_params_to_log: Optional[Iterable[str]] = None,
    api_calls: Optional[
        List[
            Union[
                Dict[str, Union[Callable[..., Any], str]],
                Union[Callable[..., Any], str],
            ]
        ]
    ] = None,
    custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
) -> Callable[[Callable[_Args, _ReturnValue]], Callable[_Args, _ReturnValue]]:
    """
    Decorator that adds function usage statement parameters to the dataframe returned by the function.

    Args:
        project: Project.
        subproject: Subproject.
        function_category: Function category.
        func_params_to_log: Function parameters to log.
        api_calls: API calls in the function.
        custom_tags: Custom tags.

    Returns:
        Decorator that adds function usage statement parameters to the dataframe returned by the decorated function.
    """

    def decorator(func: Callable[_Args, _ReturnValue]) -> Callable[_Args, _ReturnValue]:
        @functools.wraps(func)
        def wrap(*args: Any, **kwargs: Any) -> _ReturnValue:
            params = _get_func_params(func, func_params_to_log, args, kwargs) if func_params_to_log else None
            statement_params = get_function_usage_statement_params(
                project=project,
                subproject=subproject,
                function_category=function_category,
                function_name=_get_full_func_name(func),
                function_parameters=params,
                api_calls=api_calls,
                custom_tags=custom_tags,
            )

            try:
                res = func(*args, **kwargs)
                if isinstance(res, dataframe.DataFrame):
                    if hasattr(res, "_statement_params") and res._statement_params:
                        res._statement_params.update(statement_params)
                    else:
                        res._statement_params = statement_params  # type: ignore[assignment]
            except Exception:
                raise
            else:
                return res

        return cast(Callable[_Args, _ReturnValue], wrap)

    return decorator


def _get_full_func_name(func: Callable[..., Any]) -> str:
    """
    Get the full function name with module and qualname.

    Args:
        func: Function.

    Returns:
        Full function name.
    """
    module = func.__module__ if hasattr(func, "__module__") else None
    qualname = func.__qualname__
    func_name = f"{module}.{qualname}" if module else qualname
    return func_name


def _get_func_params(
    func: Callable[..., Any], func_params_to_log: Optional[Iterable[str]], args: Any, kwargs: Any
) -> Dict[str, Any]:
    """
    Get function parameters.

    Args:
        func: Function.
        func_params_to_log: Function parameters to log..
        args: Arguments.
        kwargs: Keyword arguments.

    Returns:
        Function parameters.
    """
    params = {}
    if func_params_to_log:
        spec = inspect.getfullargspec(func)
        for field in func_params_to_log:
            found, extracted_value = _extract_arg_value(field, spec, args, kwargs)
            if not found:
                pass
            else:
                params[field] = repr(extracted_value)
    return params


def _extract_arg_value(field: str, func_spec: inspect.FullArgSpec, args: Any, kwargs: Any) -> Tuple[bool, Any]:
    """
    Function to extract a specified argument value.

    Args:
        field: Target function argument name to extract.
        func_spec: Full argument spec for the function.
        args: `args` for the invoked function.
        kwargs: `kwargs` for the invoked function.

    Returns:
        Tuple: First value indicates if `field` exists.
        Second value is the extracted value if existed.
    """
    if field in func_spec.args:
        idx = func_spec.args.index(field)
        if idx < len(args):
            return True, args[idx]
        elif field in kwargs:
            return True, kwargs[field]
        else:
            if func_spec.defaults:
                required_len = len(func_spec.args) - len(func_spec.defaults)
                return True, func_spec.defaults[idx - required_len]
            return False, None
    elif func_spec.kwonlydefaults and field in func_spec.kwonlyargs:
        if field in kwargs:
            return True, kwargs[field]
        else:
            return True, func_spec.kwonlydefaults[field]
    else:
        return False, None


class _SourceTelemetryClient:
    def __init__(
        self,
        conn: connector.SnowflakeConnection,
        project: Optional[str] = None,
        subproject: Optional[str] = None,
    ) -> None:
        """
        Universal telemetry client for the source using Python connector TelemetryClient.

        Args:
            conn: SnowflakeConnection.
            project: Project.
            subproject: Subproject.

        Attributes:
            _telemetry: Python connector TelemetryClient.
            source: Source.
            version: Library version.
            python_version: Python version.
            os: Operating system.
        """
        # TODO(hayu): [SNOW-750111] Support telemetry when libraries are used in SProc.
        self._telemetry: Optional[connector_telemetry.TelemetryClient] = (
            None if utils.is_in_stored_procedure() else conn._telemetry  # type: ignore[no-untyped-call]
        )
        self.source: str = env.SOURCE
        self.project: Optional[str] = project
        self.subproject: Optional[str] = subproject
        self.version = env.VERSION
        self.python_version: str = env.PYTHON_VERSION
        self.os: str = env.OS

    def _send(self, msg: Dict[str, Any], timestamp: Optional[int] = None) -> None:
        """
        Add telemetry data to a batch in connector client.

        Args:
            msg: Telemetry message.
            timestamp: Timestamp.
        """
        if self._telemetry:
            if not timestamp:
                timestamp = time_util.get_time_millis()
            telemetry_data = connector_telemetry.TelemetryData(message=msg, timestamp=timestamp)
            self._telemetry.try_add_log_to_batch(telemetry_data)

    def _create_basic_telemetry_data(self, telemetry_type: str) -> Dict[str, Any]:
        message = {
            connector_telemetry.TelemetryField.KEY_SOURCE.value: self.source,
            TelemetryField.KEY_PROJECT.value: self.project,
            TelemetryField.KEY_SUBPROJECT.value: self.subproject,
            TelemetryField.KEY_VERSION.value: self.version,
            TelemetryField.KEY_PYTHON_VERSION.value: self.python_version,
            TelemetryField.KEY_OS.value: self.os,
            connector_telemetry.TelemetryField.KEY_TYPE.value: telemetry_type,
        }
        return message

    @suppress_exceptions
    def send_function_usage_telemetry(
        self,
        func_name: str,
        function_category: str,
        func_params: Optional[Dict[str, Any]] = None,
        api_calls: Optional[List[Dict[str, Any]]] = None,
        sfqids: Optional[List[Any]] = None,
        custom_tags: Optional[Dict[str, Union[bool, int, str, float]]] = None,
        error: Optional[str] = None,
        error_code: Optional[str] = None,
    ) -> None:
        """
        Send function usage telemetry message.

        Args:
            func_name: Function name.
            function_category: Function category.
            func_params: Function parameters.
            api_calls: API calls.
            sfqids: Snowflake query IDs.
            custom_tags: Custom tags.
            error: Error.
            error_code: Error code.
        """
        data: Dict[str, Any] = {
            TelemetryField.KEY_FUNC_NAME.value: func_name,
            TelemetryField.KEY_CATEGORY.value: function_category,
        }
        if func_params:
            data[TelemetryField.KEY_FUNC_PARAMS.value] = func_params
        if api_calls:
            data[TelemetryField.KEY_API_CALLS.value] = api_calls
        if sfqids:
            data[TelemetryField.KEY_SFQIDS.value] = sfqids
        if custom_tags:
            data[TelemetryField.KEY_CUSTOM_TAGS.value] = custom_tags

        telemetry_type = f"{self.source.lower()}_{TelemetryField.TYPE_FUNCTION_USAGE.value}"
        message: Dict[str, Any] = {
            **self._create_basic_telemetry_data(telemetry_type),
            TelemetryField.KEY_DATA.value: data,
        }

        if error:
            message[TelemetryField.KEY_ERROR_INFO.value] = error
            message[TelemetryField.KEY_ERROR_CODE.value] = error_code

        self._send(message)

    @suppress_exceptions
    def send_batch(self) -> None:
        """Send the telemetry data batch immediately."""
        if self._telemetry:
            self._telemetry.send_batch()



from typing import Optional, Union, cast

from snowflake import snowpark
from snowflake.snowpark import context, functions

CORTEX_FUNCTIONS_TELEMETRY_PROJECT = "CortexFunctions"


# Calls a sql function, handling both immediate (e.g. python types) and batch
# (e.g. snowpark column and literal type modes).
def call_sql_function(
    function: str, session: Optional[snowpark.Session], *args: Union[str, snowpark.Column]
) -> Union[str, snowpark.Column]:
    handle_as_column = False
    for arg in args:
        if isinstance(arg, snowpark.Column):
            handle_as_column = True

    if handle_as_column:
        return cast(Union[str, snowpark.Column], call_sql_function_column(function, *args))
    return cast(Union[str, snowpark.Column], call_sql_function_immediate(function, session, *args))


def call_sql_function_column(function: str, *args: Union[str, snowpark.Column]) -> snowpark.Column:
    return cast(snowpark.Column, functions.builtin(function)(*args))


def call_sql_function_immediate(
    function: str, session: Optional[snowpark.Session], *args: Union[str, snowpark.Column]
) -> str:
    if session is None:
        session = context.get_active_session()
    if session is None:
        raise Exception("No session available in the current context nor specified as an argument.")

    lit_args = []
    for arg in args:
        lit_args.append(functions.lit(arg))

    empty_df = session.create_dataframe([snowpark.Row()])
    df = empty_df.select(functions.builtin(function)(*lit_args))
    return cast(str, df.collect()[0][0])



from typing import Any, Dict, List, Optional, Tuple

from snowflake import snowpark
from snowflake.ml._internal.utils import formatting, query_result_checker

"""Table_manager is a set of utils that helps create tables.

TODO: We should make table manager a class and then put the following functions as public methods.
      Class constructor should take the session. Potentially db, schema as well.
"""


def get_fully_qualified_schema_name(database_name: str, schema_name: str) -> str:
    return f"{database_name}.{schema_name}"


def get_fully_qualified_table_name(database_name: str, schema_name: str, table_name: str) -> str:
    return f"{get_fully_qualified_schema_name(database_name, schema_name)}.{table_name}"


def create_single_table(
    session: snowpark.Session,
    database_name: str,
    schema_name: str,
    table_name: str,
    table_schema: List[Tuple[str, str]],
    statement_params: Optional[Dict[str, Any]] = None,
) -> str:
    """Creates a single table for registry and returns the fully qualified name of the table.

    Args:
        session: Session object to communicate with Snowflake.
        database_name: Desired name of the model registry database.
        schema_name: Desired name of the schema used by this model registry inside the database.
        table_name: Name of the target table.
        table_schema: A list of pair of strings, each pair denotes `(<column name>, <column type>)`.
        statement_params: Function usage statement parameters used in sql query executions.

    Returns:
        A string which is the name of the created table.

    Raises:
        RuntimeError: If table creation failed.
    """
    fully_qualified_table_name = get_fully_qualified_table_name(database_name, schema_name, table_name)
    table_schema_string = ", ".join([f"{k} {v}" for k, v in table_schema])
    try:
        session.sql(f"CREATE TABLE IF NOT EXISTS {fully_qualified_table_name} ({table_schema_string})").collect(
            statement_params=statement_params
        )
    except Exception as e:
        raise RuntimeError(f"Registry table {fully_qualified_table_name} creation failed due to {e}")

    return fully_qualified_table_name


def insert_table_entry(session: snowpark.Session, table: str, columns: Dict[str, Any]) -> List[snowpark.Row]:
    """Insert an entry into an internal Model Registry table.

    Args:
        session: Snowpark session object to communicate with Snowflake.
        table: Fully qualified name of the table to insert into.
        columns: Key-value pairs of columns and values to be inserted into the table.

    Returns:
        Result of the operation as returned by the Snowpark session (snowpark.DataFrame).

    Raises:
        RuntimeError: If entry insertion failed.
    """
    sorted_columns = sorted(columns.items())
    try:
        sql = "INSERT INTO {table} ( {columns} ) SELECT {values}".format(
            table=table,
            columns=",".join([x[0] for x in sorted_columns]),
            values=",".join([formatting.format_value_for_select(x[1]) for x in sorted_columns]),
        )
        return query_result_checker.SqlResultValidator(session, sql).insertion_success(expected_num_rows=1).validate()
    except Exception as e:
        raise RuntimeError(f"Table {table} entry {columns} insertion failed due to {e}")


def validate_table_exist(session: snowpark.Session, table: str, qualified_schema_name: str) -> bool:
    """Check if the given table exists in the target schema.

    Note:
        In case the table doesn't exist, a DataError will be raised by SqlResultValidator.

    Args:
        session: Snowpark session object to communicate with Snowflake.
        table: Name of the target table as an identifier.
        qualified_schema_name: Fully qualidied schema name where the target table is expected to exist.

    Returns:
        A boolean stands for whether the target table already exists.
    """
    tables = session.sql(f"SHOW TABLES LIKE '{table}' IN {qualified_schema_name}").collect()
    return len(tables) == 1


def get_table_schema(session: snowpark.Session, table_name: str, qualified_schema_name: str) -> Dict[str, str]:
    result = session.sql(f"DESC TABLE {qualified_schema_name}.{table_name}").collect()
    schema_dict: Dict[str, str] = {}
    for row in result:
        schema_dict[row["name"]] = row["type"]
    return schema_dict



import copy
from abc import abstractmethod
from typing import Any, Dict, Protocol, final

from snowflake.ml._internal import migrator_utils


class _BaseModelMetaMigratorProtocol(Protocol):
    source_version: str
    target_version: str

    @staticmethod
    @abstractmethod
    def upgrade(original_meta_dict: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError


class BaseModelMetaMigrator(_BaseModelMetaMigratorProtocol):
    @final
    def try_upgrade(self, original_meta_dict: Dict[str, Any]) -> Dict[str, Any]:
        loaded_meta_version = original_meta_dict.get("version", None)
        if not loaded_meta_version or str(loaded_meta_version) != self.source_version:
            raise NotImplementedError(
                f"Unknown or unsupported model metadata file with version {loaded_meta_version} found."
            )
        try:
            return self.upgrade(copy.deepcopy(original_meta_dict))
        except migrator_utils.UnableToUpgradeError as e:
            raise RuntimeError(
                f"Can not upgrade your model metadata from version {self.__class__.source_version} to"
                f" {self.__class__.target_version}."
                f"The latest version support the original version of Snowpark ML library is {e.last_supported_version}."
            )



from abc import ABC, abstractmethod
from typing import Final, Generic, Literal, Sequence

import pandas as pd
from typing_extensions import TypeGuard

from snowflake.ml.model import type_hints as model_types
from snowflake.ml.model._signatures import core


class BaseDataHandler(ABC, Generic[model_types._DataType]):
    FEATURE_PREFIX: Final[str] = "feature"
    INPUT_PREFIX: Final[str] = "input"
    OUTPUT_PREFIX: Final[str] = "output"
    SIG_INFER_ROWS_COUNT_LIMIT: Final[int] = 10

    @staticmethod
    @abstractmethod
    def can_handle(data: model_types.SupportedDataType) -> TypeGuard[model_types._DataType]:
        ...

    @staticmethod
    @abstractmethod
    def count(data: model_types._DataType) -> int:
        ...

    @staticmethod
    @abstractmethod
    def truncate(data: model_types._DataType) -> model_types._DataType:
        ...

    @staticmethod
    @abstractmethod
    def validate(data: model_types._DataType) -> None:
        ...

    @staticmethod
    @abstractmethod
    def infer_signature(
        data: model_types._DataType, role: Literal["input", "output"]
    ) -> Sequence[core.BaseFeatureSpec]:
        ...

    @staticmethod
    @abstractmethod
    def convert_to_df(data: model_types._DataType, ensure_serializable: bool = True) -> pd.DataFrame:
        ...



from typing import Optional, Union

from snowflake import snowpark
from snowflake.cortex._util import CORTEX_FUNCTIONS_TELEMETRY_PROJECT, call_sql_function
from snowflake.ml._internal import telemetry


@snowpark._internal.utils.experimental(version="1.0.12")
@telemetry.send_api_usage_telemetry(
    project=CORTEX_FUNCTIONS_TELEMETRY_PROJECT,
)
def Sentiment(
    text: Union[str, snowpark.Column], session: Optional[snowpark.Session] = None
) -> Union[str, snowpark.Column]:
    """Sentiment calls into the LLM inference service to perform sentiment analysis on the input text.

    Args:
        text: A Column of text strings to send to the LLM.
        session: The snowpark session to use. Will be inferred by context if not specified.

    Returns:
        A column of floats. 1 represents positive sentiment, -1 represents negative sentiment.
    """

    return _sentiment_impl("snowflake.ml.sentiment", text, session=session)


def _sentiment_impl(
    function: str, text: Union[str, snowpark.Column], session: Optional[snowpark.Session] = None
) -> Union[str, snowpark.Column]:
    return call_sql_function(function, session, text)



"""
- *XXX: category
    - 0XXX: undefined
    - 1XXX: system
    - 2XXX: user
    - 9XXX: internal test
- X*XX: source
    - X0XX: undefined
    - X1XX: Python built-in
    - X2XX: snowml (e.g. FileSetError)
    - X3XX: Snowpark (e.g. SnowparkClientException)
    - X4XX: Python connector (e.g. DatabaseError)
    - X5XX: Snowflake API (e.g. HTTP Error with Snowflake Image Repo)
    - X9XX: 3p dependency
- XX**: cause
"""

# INTERNAL
# Indicates an intentional error for internal error handling testing.
INTERNAL_TEST = "9000"

# UNDEFINED
# Indicates a failure that is not raised by Snowpark ML is caught by telemetry, and therefore undefined to the library,
# which can be caused by dependency APIs, unknown internal errors, etc.
UNDEFINED = "0000"

# SYSTEM
# Indicates an internal failure raising a Python built-in error with an ambiguous cause, such as invoking an unexpected
# private API, catching an error with an unknown cause, etc.
INTERNAL_PYTHON_ERROR = "1100"
# Indicates an internal failure raising a Snowpark ML error with an ambiguous cause, such as invoking an unexpected
# private API, catching an error with an unknown cause, etc.
INTERNAL_SNOWML_ERROR = "1200"
# Indicates an internal failure raising a Snowpark error with an ambiguous cause, such as invalid queries, invalid
# permission, catching an error with an unknown cause, etc.
INTERNAL_SNOWPARK_ERROR = "1300"
# Indicates an internal failure raising a error when using SPCS with an ambiguous cause, such as invalid queries,
# invalid permission, catching an error with an unknown cause, etc.
INTERNAL_SNOWPARK_CONTAINER_SERVICE_ERROR = "1301"
# Indicates an internal failure raising a Snowflake API error with an ambiguous cause, such as invalid queries, invalid
# permission, catching an error with an unknown cause, etc.
INTERNAL_SNOWFLAKE_API_ERROR = "1500"
# Indicates an internal HTTP or non-HTTP failure raising a error when interacting with Snowflake Container Services
INTERNAL_SNOWFLAKE_IMAGE_REGISTRY_ERROR = "1501"

# Indicates an internal failure raising a error when using docker with an ambiguous cause, such as invalid queries,
# invalid permission, catching an error with an unknown cause, etc.
INTERNAL_DOCKER_ERROR = "1901"

# USER
# Indicates the incompatibility of local dependency versions with the target requirements. For example, an API added in
# a later version is called with an older dependency installed.
DEPENDENCY_VERSION_ERROR = "2100"
# Indicates the resource is missing: not whether the absence is temporary or permanent.
NOT_FOUND = "2101"
# The method is known but is not supported by the target resource. For example, calling `to_xgboost` is not allowed by
# Snowpark ML models based on scikit-learn.
METHOD_NOT_ALLOWED = "2102"
# Not implemented.
NOT_IMPLEMENTED = "2103"

# Calling an API with unsupported keywords/values.
INVALID_ARGUMENT = "2110"
# Object has invalid attributes caused by invalid/unsupported value, unsupported data type, size mismatch, etc.
INVALID_ATTRIBUTE = "2111"
# Missing and invalid data caused by null value, unexpected value (e.g. division by 0), out of range value, etc.
INVALID_DATA = "2112"
# Invalid data type in the processed data. For example, an API handling numeric columns gets a string column.
INVALID_DATA_TYPE = "2113"
# Calling an API with unsupported value type, or perform actions on objects with incorrect types.
INVALID_TYPE = "2114"
# Trying to create an object already exists.
OBJECT_ALREADY_EXISTS = "2115"

# Indicates the creation of underlying resources (files, stages, tables, etc) failed, which can be caused by duplicated
# name, invalid permission, etc.
SNOWML_CREATE_FAILED = "2200"
# Indicates the read of underlying resources (files, stages, tables, etc) failed, which can be caused by duplicated
# name, invalid permission, etc.
SNOWML_READ_FAILED = "2201"
# Indicates the update of underlying resources (files, stages, tables, etc) failed, which can be caused by duplicated
# name, invalid permission, etc.
SNOWML_UPDATE_FAILED = "2202"
# Indicates the deletion of underlying resources (files, stages, tables, etc) failed, which can be caused by duplicated
# name, invalid permission, etc.
SNOWML_DELETE_FAILED = "2203"
# Indicates the Snowflake resource is missing: not whether the absence is temporary or permanent.
SNOWML_NOT_FOUND = "2204"
# Indicates the access of a stage failed, which can be caused by invalid name, invalid permission, etc.
SNOWML_INVALID_STAGE = "2210"
# Invalid query caused by syntax error, invalid source, etc.
SNOWML_INVALID_QUERY = "2211"

# Invalid Snowpark Session (Missing information) in Snowpark Session that is required.
INVALID_SNOWPARK_SESSION = "2301"
# Indicates the compute pool is not valid, e.g. not ready, or does not have enough GPUs.
INVALID_SNOWPARK_COMPUTE_POOL = "2302"

# Incorrect local Python environment when trying to do some actions that require specific dependency or versions.
LOCAL_ENVIRONMENT_ERROR = "2501"
# Unfeasible dependencies requirement when trying to do some actions that require specific environments.
UNFEASIBLE_ENVIRONMENT_ERROR = "2502"

# Missing required client side dependency.
CLIENT_DEPENDENCY_MISSING_ERROR = "2511"



_KEEP_ORDER_COL_NAME = "_ID"

_UDF_CODE_TEMPLATE = """
import fcntl
import functools
import inspect
import os
import sys
import threading
import zipfile
from types import TracebackType
from typing import Optional, Type

import anyio
import pandas as pd
from _snowflake import vectorized


class FileLock:
    def __enter__(self) -> None:
        self._lock = threading.Lock()
        self._lock.acquire()
        self._fd = open("/tmp/lockfile.LOCK", "w+")
        fcntl.lockf(self._fd, fcntl.LOCK_EX)

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], traceback: Optional[TracebackType]
    ) -> None:
        self._fd.close()
        self._lock.release()


# User-defined parameters
MODEL_FILE_NAME = "{model_stage_file_name}"
TARGET_METHOD = "{target_method}"
MAX_BATCH_SIZE = None


# Retrieve the model
IMPORT_DIRECTORY_NAME = "snowflake_import_directory"
import_dir = sys._xoptions[IMPORT_DIRECTORY_NAME]

model_dir_name = os.path.splitext(MODEL_FILE_NAME)[0]
zip_model_path = os.path.join(import_dir, MODEL_FILE_NAME)
extracted = "/tmp/models"
extracted_model_dir_path = os.path.join(extracted, model_dir_name)

with FileLock():
    if not os.path.isdir(extracted_model_dir_path):
        with zipfile.ZipFile(zip_model_path, "r") as myzip:
            myzip.extractall(extracted_model_dir_path)

sys.path.insert(0, os.path.join(extracted_model_dir_path, "{code_dir_name}"))

# Load the model
try:
    from snowflake.ml.model._packager import model_packager
    pk = model_packager.ModelPackager(extracted_model_dir_path)
    pk.load(as_custom_model=True)
    assert pk.model, "model is not loaded"
    assert pk.meta, "model metadata is not loaded"

    model = pk.model
    meta = pk.meta
except ImportError as e:
    if e.name and not e.name.startswith("snowflake.ml"):
        raise e
    # Support Legacy model
    from snowflake.ml.model import _model
    # Backward for <= 1.0.5
    if hasattr(_model, "_load_model_for_deploy"):
        model, meta = _model._load_model_for_deploy(extracted_model_dir_path)
    else:
        model, meta = _model._load(local_dir_path=extracted_model_dir_path, as_custom_model=True)

# Determine the actual runner
func = getattr(model, TARGET_METHOD)
if inspect.iscoroutinefunction(func):
    runner = functools.partial(anyio.run, func)
else:
    runner = functools.partial(func)

# Determine preprocess parameters
features = meta.signatures[TARGET_METHOD].inputs
input_cols = [feature.name for feature in features]
dtype_map = {{feature.name: feature.as_dtype() for feature in features}}


# Actual handler
@vectorized(input=pd.DataFrame, max_batch_size=MAX_BATCH_SIZE)
def infer(df: pd.DataFrame) -> dict:
    input_df = pd.json_normalize(df[0]).astype(dtype=dtype_map)
    predictions_df = runner(input_df[input_cols])

    if "{_KEEP_ORDER_COL_NAME}" in input_df.columns:
        predictions_df["{_KEEP_ORDER_COL_NAME}"] = input_df["{_KEEP_ORDER_COL_NAME}"]

    return predictions_df.to_dict("records")
"""



import http
import logging
from typing import Dict, Optional, cast
from urllib.parse import urlunparse

from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions as snowml_exceptions,
)
from snowflake.ml._internal.utils import image_registry_http_client
from snowflake.ml.model._deploy_client.utils import imagelib
from snowflake.snowpark import Session
from snowflake.snowpark._internal import utils as snowpark_utils

_MANIFEST_V1_HEADER = "application/vnd.oci.image.manifest.v1+json"
_MANIFEST_V2_HEADER = "application/vnd.docker.distribution.manifest.v2+json"
_SUPPORTED_MANIFEST_HEADERS = [_MANIFEST_V1_HEADER, _MANIFEST_V2_HEADER]

logger = logging.getLogger(__name__)


class ImageRegistryClient:
    """
    A partial implementation of an SPCS image registry client. The client utilizes the ImageRegistryHttpClient under
    the hood, incorporating a retry mechanism to handle intermittent 401 errors from the SPCS image registry.
    """

    def __init__(self, session: Session, full_dest_image_name: str) -> None:
        """Initialization

        Args:
            session: Snowpark session
            full_dest_image_name: Based on dest image name, repo url can be inferred.
        """
        self.image_registry_http_client = image_registry_http_client.ImageRegistryHttpClient(
            session=session,
            repo_url=self._convert_to_v2_manifests_url(full_image_name=full_dest_image_name),
        )

    def _convert_to_v2_manifests_url(self, full_image_name: str) -> str:
        """Converts a full image name to a Docker Registry HTTP API V2 URL:
        https://docs.docker.com/registry/spec/api/#existing-manifests

        org-account.registry-dev.snowflakecomputing.com/db/schema/repo/image_name:tag becomes
        https://org-account.registry-dev.snowflakecomputing.com/v2/db/schema/repo/image_name/manifests/tag

        Args:
            full_image_name: a string consists of image name and image tag.

        Returns:
            Docker HTTP V2 URL for checking manifest existence.
        """
        scheme = "https"
        full_image_name_parts = full_image_name.split(":")
        assert len(full_image_name_parts) == 2, "full image name should include both image name and tag"

        image_name = full_image_name_parts[0]
        tag = full_image_name_parts[1]
        image_name_parts = image_name.split("/")
        domain = image_name_parts[0]
        rest = "/".join(image_name_parts[1:])
        path = f"/v2/{rest}/manifests/{tag}"
        url_tuple = (scheme, domain, path, "", "", "")
        return urlunparse(url_tuple)

    def _get_accept_headers(self) -> Dict[str, str]:
        # Depending on the built image, the media type of the image manifest might be either
        # application/vnd.oci.image.manifest.v1+json or application/vnd.docker.distribution.manifest.v2+json
        # Hence we need to check for both, otherwise it could result in false negative.
        return {"Accept": ",".join(_SUPPORTED_MANIFEST_HEADERS)}

    def image_exists(self, full_image_name: str) -> bool:
        """Check whether image already exists in the registry.

        Args:
            full_image_name: Full image name consists of image name and image tag.

        Returns:
            Boolean value. True when image already exists, else False.

        """
        # When running in SPROC, the Sproc session connection will not have _rest object associated, which makes it
        # unable to fetch session token needed to authenticate to SPCS image registry.
        if snowpark_utils.is_in_stored_procedure():  # type: ignore[no-untyped-call]
            return False
        v2_api_url = self._convert_to_v2_manifests_url(full_image_name)
        headers = self._get_accept_headers()
        status = self.image_registry_http_client.head(v2_api_url, headers=headers).status_code
        return status == http.HTTPStatus.OK

    def _get_manifest(self, full_image_name: str) -> Dict[str, str]:
        """Retrieve image manifest file. Given Docker manifest comes with two versions, and for each version the
        corresponding request header is required for a successful HTTP response.

        Args:
            full_image_name: Full image name.

        Returns:
            Full manifest content as a python dict.

        Raises:
            SnowflakeMLException: when failed to retrieve manifest.
        """

        v2_api_url = self._convert_to_v2_manifests_url(full_image_name)
        res = self.image_registry_http_client.get(v2_api_url, headers=self._get_accept_headers())
        if res.status_code == http.HTTPStatus.OK:
            return cast(Dict[str, str], res.json())
        raise snowml_exceptions.SnowflakeMLException(
            error_code=error_codes.INTERNAL_SNOWFLAKE_IMAGE_REGISTRY_ERROR,
            original_exception=ValueError(
                f"Failed to retrieve manifest for {full_image_name}. \n"
                f"HTTP status code: {res.status_code}. Full response: {res.text}."
            ),
        )

    def add_tag_to_remote_image(self, original_full_image_name: str, new_tag: str) -> None:
        """Add a tag to an image in the registry.

        Args:
            original_full_image_name: The full image name is required to fetch manifest.
            new_tag:  New tag to be added to the image.

        Returns:
            None

        Raises:
            SnowflakeMLException: when failed to push the newly updated manifest.
        """

        if snowpark_utils.is_in_stored_procedure():  # type: ignore[no-untyped-call]
            return None

        full_image_name_parts = original_full_image_name.split(":")
        assert len(full_image_name_parts) == 2, "full image name should include both image name and tag"
        new_full_image_name = ":".join([full_image_name_parts[0], new_tag])
        if self.image_exists(new_full_image_name):
            # Early return if image with the associated tag already exists.
            return
        api_url = self._convert_to_v2_manifests_url(new_full_image_name)
        manifest = self._get_manifest(full_image_name=original_full_image_name)
        manifest_copy = manifest.copy()
        manifest_copy["tag"] = new_tag
        headers = self._get_accept_headers()
        # Http Content-Type does not support multi-value, hence need to construct separate header.
        put_header_v1 = {
            **headers,
            "Content-Type": _MANIFEST_V1_HEADER,
        }
        put_header_v2 = {
            **headers,
            "Content-Type": _MANIFEST_V2_HEADER,
        }

        res1 = self.image_registry_http_client.put(api_url, headers=put_header_v1, json=manifest_copy)
        if res1.status_code != http.HTTPStatus.CREATED:
            res2 = self.image_registry_http_client.put(api_url, headers=put_header_v2, json=manifest_copy)
            if res2.status_code != http.HTTPStatus.CREATED:
                raise snowml_exceptions.SnowflakeMLException(
                    error_code=error_codes.INTERNAL_SNOWFLAKE_IMAGE_REGISTRY_ERROR,
                    original_exception=ValueError(
                        f"Failed to push manifest for {new_full_image_name}. Two requests filed: \n"
                        f"HTTP status code 1: {res1.status_code}. Full response 1: {res1.text}. \n"
                        f"HTTP status code 2: {res2.status_code}. Full response 2: {res2.text}"
                    ),
                )
        assert self.image_exists(
            new_full_image_name
        ), f"{new_full_image_name} should exist in image repo after a successful manifest update"

    def copy_image(
        self,
        source_image_with_digest: str,
        dest_image_with_tag: str,
        arch: Optional[imagelib._Arch] = None,
    ) -> None:
        """Util function to copy image across registry. Currently supported pulling from public docker image repo to
        SPCS image registry.

        Args:
            source_image_with_digest: source image with digest, e.g. gcr.io/kaniko-project/executor@sha256:b8c0977
            dest_image_with_tag: destination image with tag.
            arch: architecture of source image.

        Returns:
            None
        """
        logger.info(f"Copying image from {source_image_with_digest} to {dest_image_with_tag}")
        if snowpark_utils.is_in_stored_procedure():  # type: ignore[no-untyped-call]
            logger.warning(f"Running inside Sproc. Please ensure image already exists at {dest_image_with_tag}")
            return None

        arch = arch or imagelib._Arch("amd64", "linux")

        src_image = imagelib.convert_to_image_descriptor(source_image_with_digest, with_digest=True)
        dest_image = imagelib.convert_to_image_descriptor(
            dest_image_with_tag,
            with_tag=True,
        )
        # TODO[shchen]: Remove the imagelib, instead rely on the copy image system function later.
        imagelib.copy_image(
            src_image=src_image, dest_image=dest_image, arch=arch, retryable_http=self.image_registry_http_client
        )
        logger.info("Image copy completed successfully")



from typing import Dict, Optional, Tuple

from typing_extensions import TypedDict

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.snowpark import functions

_PROJECT = "MLOps"
_SUBPROJECT = "Monitor"


class BucketConfig(TypedDict):
    """ "Options for bucketizing the data."""

    min: int
    max: int
    size: int


@telemetry.send_api_usage_telemetry(
    project=_PROJECT,
    subproject=_SUBPROJECT,
)
@snowpark._internal.utils.private_preview(version="1.0.10")  # TODO: update versions when release
def compare_udfs_outputs(
    base_udf_name: str,
    test_udf_name: str,
    input_data_df: snowpark.DataFrame,
    bucket_config: Optional[BucketConfig] = None,
) -> snowpark.DataFrame:
    """Compare outputs of 2 UDFs. Outputs are bucketized the based on bucketConfig.
    This is useful when someone retrain a Model and deploy as UDF to compare against earlier UDF as ground truth.
    NOTE: Only supports UDFs with single Column output.

    Args:
        base_udf_name: used as control ground truth UDF.
        test_udf_name: output of this UDF is compared against that of `base_udf`.
        input_data_df: Input data used for computing metric.
        bucket_config: must have the kv as {"min":xx, "max":xx, "size"}, keys in lowercase; it's width_bucket
            Sqloperator's config, using https://docs.snowflake.com/en/sql-reference/functions/width_bucket.

    Returns:
        snowpark.DataFrame.
        "BASEUDF" is base_udf's bucketized output, "TESTUDF" is test_udf's bucketized output,
    """
    if bucket_config:
        assert len(bucket_config) == 3
        assert "min" in bucket_config and "max" in bucket_config and "size" in bucket_config

    argStr = ",".join(input_data_df.columns)
    query1Str = _get_udf_query_str("BASEUDF", f"{base_udf_name}({argStr})", input_data_df, bucket_config)
    query2Str = _get_udf_query_str("TESTUDF", f"{test_udf_name}({argStr})", input_data_df, bucket_config)

    if bucket_config:
        finalStr = (
            "select A.bucket, BASEUDF, TESTUDF \n from ({}) as A \n join ({}) as B \n on A.bucket=B.bucket".format(
                query1Str, query2Str
            )
        )
    else:  # don't bucket at all
        finalStr = "select {},{} \n from ({})".format(query1Str, query2Str, input_data_df.queries["queries"][0])

    assert input_data_df._session is not None
    return input_data_df._session.sql(finalStr)


@telemetry.send_api_usage_telemetry(
    project=_PROJECT,
    subproject=_SUBPROJECT,
)
@snowpark._internal.utils.private_preview(version="1.0.10")  # TODO: update versions when release
def get_basic_stats(df: snowpark.DataFrame) -> Tuple[Dict[str, int], Dict[str, int]]:
    """Get basic stats of 2 Columns
    Note this isn't public API. Only support min, max, stddev, HLL--cardinality estimate

    Args:
        df: input Snowpark Dataframe, must have 2 and only 2 columns

    Returns:
        2 Dict for 2 columns' stats
    """
    projStr = ""
    stats = ["MIN", "MAX", "STDDEV", "HLL"]
    assert len(df.columns) == 2
    for colName in df.columns:
        for stat in stats:
            projStr += f"{stat}({colName}) as {colName}_{stat},"
    finalStr = "select {} \n from ({})".format(projStr[:-1], df.queries["queries"][0])
    assert df._session is not None
    resDf = df._session.sql(finalStr).to_pandas()
    d1 = {}
    col1 = df.columns[0]
    d2 = {}
    col2 = df.columns[1]
    for stat in stats:
        d1[stat] = resDf.iloc[0][f"{col1}_{stat}"]
        d2[stat] = resDf.iloc[0][f"{col2}_{stat}"]
    return d1, d2


@telemetry.send_api_usage_telemetry(
    project=_PROJECT,
    subproject=_SUBPROJECT,
)
@snowpark._internal.utils.private_preview(version="1.0.10")  # TODO: update versions when release
def jensenshannon(df1: snowpark.DataFrame, colname1: str, df2: snowpark.DataFrame, colname2: str) -> float:
    """
    Similar to scipy implementation:
    https://github.com/scipy/scipy/blob/e4dec2c5993faa381bb4f76dce551d0d79734f8f/scipy/spatial/distance.py#L1174
    It's server solution, all computing being in Snowflake warehouse, so will be significantly faster than client.

    Args:
        df1: 1st Snowpark Dataframe;
        colname1: the col to be selected in df1
        df2: 2nd Snowpark Dataframe;
        colname2: the col to be selected in df2
            Supported data Tyte: any data type that Snowflake supports, including VARIANT, OBJECT...etc.

    Returns:
        a jensenshannon value
    """
    df1 = df1.select(colname1)
    df1 = (
        df1.group_by(colname1)
        .agg(functions.count(colname1).alias("c1"))
        .select(functions.col(colname1).alias("d1"), "c1")
    )
    df2 = df2.select(colname2)
    df2 = (
        df2.group_by(colname2)
        .agg(functions.count(colname2).alias("c2"))
        .select(functions.col(colname2).alias("d2"), "c2")
    )

    dfsum = df1.select("c1").agg(functions.sum("c1").alias("SUM1"))
    sum1 = dfsum.collect()[0].as_dict()["SUM1"]
    dfsum = df2.select("c2").agg(functions.sum("c2").alias("SUM2"))
    sum2 = dfsum.collect()[0].as_dict()["SUM2"]

    df1 = df1.select("d1", functions.sql_expr("c1 / " + str(sum1)).alias("p"))
    minp = df1.select(functions.min("P").alias("MINP")).collect()[0].as_dict()["MINP"]
    df2 = df2.select("d2", functions.sql_expr("c2 / " + str(sum2)).alias("q"))
    minq = df2.select(functions.min("Q").alias("MINQ")).collect()[0].as_dict()["MINQ"]

    DECAY_FACTOR = 0.5
    df = df1.join(df2, df1.d1 == df2.d2, "fullouter").select(
        "d1",
        "d2",
        functions.sql_expr(
            """
            CASE
                WHEN p is NULL THEN {}*{}
                ELSE p
            END
            """.format(
                minp, DECAY_FACTOR
            )
        ).alias("p"),
        functions.sql_expr(
            """
            CASE
                WHEN q is NULL THEN {}*{}
                ELSE q
            END
            """.format(
                minq, DECAY_FACTOR
            )
        ).alias("q"),
    )

    df = df.select("p", "q", functions.sql_expr("(p+q)/2.0").alias("m"))
    df = df.select(
        functions.sql_expr(
            """
            CASE
                WHEN p > 0 AND m > 0 THEN p * LOG(2, p/m)
                ELSE 0
            END
            """
        ).alias("left"),
        functions.sql_expr(
            """
            CASE
                WHEN q > 0 AND m > 0 THEN q * LOG(2, q/m)
                ELSE 0
            END
            """
        ).alias("right"),
    )
    resdf = df.select(functions.sql_expr("sqrt((sum(left) + sum(right)) / 2.0)").alias("JS"))
    return float(resdf.collect()[0].as_dict()["JS"])


def _get_udf_query_str(
    name: str, col: str, df: snowpark.DataFrame, bucket_config: Optional[BucketConfig] = None
) -> str:
    if bucket_config:
        return "select count(1) as {}, width_bucket({}, {}, {}, {}) bucket from ({}) group by bucket".format(
            name, col, bucket_config["min"], bucket_config["max"], bucket_config["size"], df.queries["queries"][0]
        )
    else:  # don't bucket at all
        return f"{col} as {name}"



#!/usr/bin/env python3
import inspect
from abc import abstractmethod
from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, Iterable, List, Mapping, Optional, Union, overload

import numpy as np
import numpy.typing as npt
import pandas as pd

from snowflake import snowpark
from snowflake.ml._internal import telemetry
from snowflake.ml._internal.exceptions import (
    error_codes,
    exceptions,
    modeling_error_messages,
)
from snowflake.ml._internal.utils import parallelize
from snowflake.ml.modeling.framework import _utils
from snowflake.snowpark import functions as F

PROJECT = "ModelDevelopment"
SUBPROJECT = "Preprocessing"


def _process_cols(cols: Optional[Union[str, Iterable[str]]]) -> List[str]:
    """Convert cols to a list."""
    col_list: List[str] = []
    if cols is None:
        return col_list
    elif type(cols) is list:
        col_list = cols
    elif type(cols) is str:
        col_list = [cols]
    else:
        col_list = list(cols)

    return col_list


class Base:
    def __init__(self) -> None:
        """
        Base class for all estimators and transformers.

        Attributes:
            input_cols: Input columns.
            output_cols: Output columns.
            label_cols: Label column(s).
        """
        self.input_cols: List[str] = []
        self.output_cols: List[str] = []
        self.label_cols: List[str] = []

    def _create_unfitted_sklearn_object(self) -> Any:
        raise NotImplementedError()

    def _create_sklearn_object(self) -> Any:
        raise NotImplementedError()

    def get_input_cols(self) -> List[str]:
        """
        Input columns getter.

        Returns:
            Input columns.
        """
        return self.input_cols

    def set_input_cols(self, input_cols: Optional[Union[str, Iterable[str]]]) -> "Base":
        """
        Input columns setter.

        Args:
            input_cols: A single input column or multiple input columns.

        Returns:
            self
        """
        self.input_cols = _process_cols(input_cols)
        return self

    def get_output_cols(self) -> List[str]:
        """
        Output columns getter.

        Returns:
            Output columns.
        """
        return self.output_cols

    def set_output_cols(self, output_cols: Optional[Union[str, Iterable[str]]]) -> "Base":
        """
        Output columns setter.

        Args:
            output_cols: A single output column or multiple output columns.

        Returns:
            self
        """
        self.output_cols = _process_cols(output_cols)
        return self

    def get_label_cols(self) -> List[str]:
        """
        Label column getter.

        Returns:
            Label column(s).
        """
        return self.label_cols

    def set_label_cols(self, label_cols: Optional[Union[str, Iterable[str]]]) -> "Base":
        """
        Label column setter.

        Args:
            label_cols: A single label column or multiple label columns if multi task learning.

        Returns:
            self
        """
        self.label_cols = _process_cols(label_cols)
        return self

    def _check_input_cols(self) -> None:
        """
        Check if `self.input_cols` is set.

        Raises:
            SnowflakeMLException: `self.input_cols` is not set.
        """
        if not self.input_cols:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_FOUND,
                original_exception=RuntimeError(modeling_error_messages.ATTRIBUTE_NOT_SET.format("input_cols")),
            )

    def _check_output_cols(self) -> None:
        """
        Check if `self.output_cols` is set.

        Raises:
            SnowflakeMLException: `self.output_cols` is not set.
            SnowflakeMLException: `self.output_cols` and `self.input_cols` are of different lengths.
        """
        if not self.output_cols:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_FOUND,
                original_exception=RuntimeError(modeling_error_messages.ATTRIBUTE_NOT_SET.format("output_cols")),
            )
        if len(self.output_cols) != len(self.input_cols):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=RuntimeError(
                    modeling_error_messages.SIZE_MISMATCH.format(
                        "input_cols", len(self.input_cols), "output_cols", len(self.output_cols)
                    )
                ),
            )

    @staticmethod
    def _check_dataset_type(dataset: Any) -> None:
        """
        Check if the type of input dataset is supported.

        Args:
            dataset: Input dataset passed to an API.

        Raises:
            SnowflakeMLException: `self.input_cols` is not set.
        """
        if not (isinstance(dataset, snowpark.DataFrame) or isinstance(dataset, pd.DataFrame)):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ARGUMENT,
                original_exception=TypeError(
                    f"Unexpected dataset type: {type(dataset)}."
                    f"Supported dataset types: {type(snowpark.DataFrame)}, {type(pd.DataFrame)}."
                ),
            )

    @classmethod
    def _get_param_names(cls) -> List[str]:
        """Get parameter names for the transformer"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values() if p.name != "self" and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise exceptions.SnowflakeMLException(
                    error_code=error_codes.INTERNAL_PYTHON_ERROR,
                    original_exception=RuntimeError(
                        "Models should always specify"
                        " their parameters in the signature"
                        " of their __init__ (no varargs)."
                        f" {cls} with constructor {init_signature} doesn't "
                        " follow this convention."
                    ),
                )
        # Extract and sort argument names excluding 'self'
        return sorted(p.name for p in parameters)

    def get_params(self, deep: bool = True) -> Dict[str, Any]:
        """
        Get parameters for this transformer.

        Args:
            deep: If True, will return the parameters for this transformer and
                contained subobjects that are transformers.

        Returns:
            Parameter names mapped to their values.
        """
        out: Dict[str, Any] = dict()
        for key in self._get_param_names():
            if hasattr(self, key):
                value = getattr(self, key)
                if deep and hasattr(value, "get_params"):
                    deep_items = value.get_params().items()
                    out.update((key + "__" + k, val) for k, val in deep_items)
                out[key] = value
        return out

    def set_params(self, **params: Dict[str, Any]) -> None:
        """
        Set the parameters of this transformer.

        The method works on simple transformers as well as on nested objects.
        The latter have parameters of the form ``<component>__<parameter>``
        so that it's possible to update each component of a nested object.

        Args:
            **params: Transformer parameter names mapped to their values.

        Raises:
            SnowflakeMLException: Invalid parameter keys.
        """
        if not params:
            # simple optimization to gain speed (inspect is slow)
            return
        valid_params = self.get_params(deep=True)

        nested_params: Dict[str, Any] = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                local_valid_params = self._get_param_names()
                raise exceptions.SnowflakeMLException(
                    error_code=error_codes.INVALID_ARGUMENT,
                    original_exception=ValueError(
                        modeling_error_messages.INVALID_MODEL_PARAM.format(
                            key, self.__class__.__name__, local_valid_params
                        )
                    ),
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

    def get_sklearn_args(
        self,
        default_sklearn_obj: Optional[object] = None,
        sklearn_initial_keywords: Optional[Union[str, Iterable[str]]] = None,
        sklearn_unused_keywords: Optional[Union[str, Iterable[str]]] = None,
        snowml_only_keywords: Optional[Union[str, Iterable[str]]] = None,
        sklearn_added_keyword_to_version_dict: Optional[Dict[str, str]] = None,
        sklearn_added_kwarg_value_to_version_dict: Optional[Dict[str, Dict[str, str]]] = None,
        sklearn_deprecated_keyword_to_version_dict: Optional[Dict[str, str]] = None,
        sklearn_removed_keyword_to_version_dict: Optional[Dict[str, str]] = None,
    ) -> Dict[str, Any]:
        """
        Get sklearn keyword arguments.

        This method enables modifying object parameters for special cases.

        Args:
            default_sklearn_obj: Sklearn object used to get default parameter values. Necessary when
                `sklearn_added_keyword_to_version_dict` is provided.
            sklearn_initial_keywords: Initial keywords in sklearn.
            sklearn_unused_keywords: Sklearn keywords that are unused in snowml.
            snowml_only_keywords: snowml only keywords not present in sklearn.
            sklearn_added_keyword_to_version_dict: Added keywords mapped to the sklearn versions in which they were
                added.
            sklearn_added_kwarg_value_to_version_dict: Added keyword argument values mapped to the sklearn versions
                in which they were added.
            sklearn_deprecated_keyword_to_version_dict: Deprecated keywords mapped to the sklearn versions in which
                they were deprecated.
            sklearn_removed_keyword_to_version_dict: Removed keywords mapped to the sklearn versions in which they
                were removed.

        Returns:
            Sklearn parameter names mapped to their values.
        """
        default_sklearn_args = _utils.get_default_args(default_sklearn_obj.__class__.__init__)
        given_args = self.get_params()
        sklearn_args: Dict[str, Any] = _utils.get_filtered_valid_sklearn_args(
            args=given_args,
            default_sklearn_args=default_sklearn_args,
            sklearn_initial_keywords=sklearn_initial_keywords,
            sklearn_unused_keywords=sklearn_unused_keywords,
            snowml_only_keywords=snowml_only_keywords,
            sklearn_added_keyword_to_version_dict=sklearn_added_keyword_to_version_dict,
            sklearn_added_kwarg_value_to_version_dict=sklearn_added_kwarg_value_to_version_dict,
            sklearn_deprecated_keyword_to_version_dict=sklearn_deprecated_keyword_to_version_dict,
            sklearn_removed_keyword_to_version_dict=sklearn_removed_keyword_to_version_dict,
        )
        return sklearn_args


class BaseEstimator(Base):
    def __init__(
        self,
        *,
        file_names: Optional[List[str]] = None,
        custom_states: Optional[List[str]] = None,
        sample_weight_col: Optional[str] = None,
    ) -> None:
        """
        Base class for all estimators.

        Following scikit-learn APIs, all estimators should specify all
        the parameters that can be set at the class level in their ``__init__``
        as explicit keyword arguments (no ``*args`` or ``**kwargs``)

        Args:
            file_names: File names.
            custom_states: Custom states.
            sample_weight_col: Sample weight column.

        Attributes:
            start_time: Start time of the transformer.
        """
        super().__init__()

        # transformer state
        self.file_names = file_names
        self.custom_states = custom_states
        self.sample_weight_col = sample_weight_col

        self.start_time = datetime.now().strftime(_utils.DATETIME_FORMAT)[:-3]

    def get_sample_weight_col(self) -> Optional[str]:
        """
        Sample weight column getter.

        Returns:
            Sample weight column.
        """
        return self.sample_weight_col

    def set_sample_weight_col(self, sample_weight_col: Optional[str]) -> "Base":
        """
        Sample weight column setter.

        Args:
            sample_weight_col: A single column that represents sample weight.

        Returns:
            self
        """
        self.sample_weight_col = sample_weight_col
        return self

    def _get_dependencies(self) -> List[str]:
        """
        Return the list of conda dependencies required to work with the object.

        Returns:
            List of dependencies.
        """
        return []

    @abstractmethod
    def fit(self, dataset: snowpark.DataFrame) -> "BaseEstimator":
        raise NotImplementedError()

    def _use_input_cols_only(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """Returns the pandas dataframe filtered on `input_cols`, or raises an error if malformed."""
        input_cols = set(self.input_cols)
        dataset_cols = set(dataset.columns.to_list())
        if not input_cols.issubset(dataset_cols):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.NOT_FOUND,
                original_exception=KeyError(
                    "input_cols contains columns that do not match any of the columns in "
                    f"the pandas dataframe: {input_cols - dataset_cols}."
                ),
            )
        return dataset[self.input_cols]

    def _compute(
        self, dataset: snowpark.DataFrame, cols: List[str], states: List[str]
    ) -> Dict[str, Dict[str, Union[int, float, str]]]:
        """
        Compute required states of the columns.

        Args:
            dataset: Input dataset.
            cols: Columns to compute.
            states: States to compute. Arguments are appended. For example, "min" represents
                getting min value of the column, and "percentile_cont:0.25" represents getting
                the 25th percentile disc value of the column.

        Returns:
            A dict of {column_name: {state: value}} of each column.
        """

        def _compute_on_partition(df: snowpark.DataFrame, cols_subset: List[str]) -> snowpark.DataFrame:
            """Returns a DataFrame with the desired computation on the specified column subset."""
            exprs = []
            sql_prefix = "SQL>>>"

            for col_name in cols_subset:
                for state in states:
                    if state.startswith(sql_prefix):
                        sql_expr = state[len(sql_prefix) :].format(col_name=col_name)
                        exprs.append(sql_expr)
                    else:
                        func = _utils.STATE_TO_FUNC_DICT[state].__name__
                        exprs.append(f"{func}({col_name})")

            res: snowpark.DataFrame = df.select_expr(exprs)
            return res

        _results = parallelize.map_dataframe_by_column(
            df=dataset,
            cols=cols,
            map_func=_compute_on_partition,
            partition_size=40,
            statement_params=telemetry.get_statement_params(PROJECT, SUBPROJECT, self.__class__.__name__),
        )

        computed_dict: Dict[str, Dict[str, Union[int, float, str]]] = {}
        for idx, val in enumerate(_results[0]):
            col_name = cols[idx // len(states)]
            if col_name not in computed_dict:
                computed_dict[col_name] = {}

            state = states[idx % len(states)]
            computed_dict[col_name][state] = val

        return computed_dict


class BaseTransformer(BaseEstimator):
    def __init__(
        self,
        *,
        drop_input_cols: Optional[bool] = False,
        file_names: Optional[List[str]] = None,
        custom_states: Optional[List[str]] = None,
        sample_weight_col: Optional[str] = None,
    ) -> None:
        """Base class for all transformers."""
        super().__init__(file_names=file_names, custom_states=custom_states, sample_weight_col=sample_weight_col)
        self._sklearn_object = None
        self._is_fitted = False
        self._drop_input_cols = drop_input_cols

    @overload
    def transform(self, dataset: snowpark.DataFrame) -> snowpark.DataFrame:
        ...

    @overload
    def transform(self, dataset: pd.DataFrame) -> pd.DataFrame:
        ...

    @abstractmethod
    def transform(self, dataset: Union[snowpark.DataFrame, pd.DataFrame]) -> Union[snowpark.DataFrame, pd.DataFrame]:
        raise NotImplementedError()

    def _enforce_fit(self) -> None:
        if not self._is_fitted:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.METHOD_NOT_ALLOWED,
                original_exception=RuntimeError(
                    f"Model {self.__class__.__name__} not fitted before calling predict/transform."
                ),
            )

    def set_drop_input_cols(self, drop_input_cols: Optional[bool] = False) -> None:
        self._drop_input_cols = drop_input_cols

    def to_sklearn(self) -> Any:
        if self._sklearn_object is None:
            self._sklearn_object = self._create_sklearn_object()
        return self._sklearn_object

    # to_xgboost would be only used in XGB estimators
    # to_lightgbm would be only used in LightGBM estimators, but they function the same
    def to_xgboost(self) -> Any:
        raise exceptions.SnowflakeMLException(
            error_code=error_codes.METHOD_NOT_ALLOWED,
            original_exception=AttributeError(
                modeling_error_messages.UNSUPPORTED_MODEL_CONVERSION.format("to_xgboost", "to_sklearn")
            ),
        )

    def to_lightgbm(self) -> Any:
        raise exceptions.SnowflakeMLException(
            error_code=error_codes.METHOD_NOT_ALLOWED,
            original_exception=AttributeError(
                modeling_error_messages.UNSUPPORTED_MODEL_CONVERSION.format("to_lightgbm", "to_sklearn")
            ),
        )

    def _reset(self) -> None:
        self._sklearn_object = None
        self._is_fitted = False

    def _convert_attribute_dict_to_ndarray(
        self,
        attribute: Optional[Mapping[str, Union[int, float, str, Iterable[Union[int, float, str]]]]],
        dtype: Optional[type] = None,
    ) -> Optional[npt.NDArray[Union[np.int_, np.float_, np.str_]]]:
        """
        Convert the attribute from dict to ndarray based on the order of `self.input_cols`.

        Args:
            attribute: Attribute to convert. Attribute is a mapping of a column to its attribute value(s)
                (e.g. `StandardScaler.mean_: {column_name: mean_value}`).
            dtype: The dtype of the converted ndarray. If None, there is no type conversion.

        Returns:
            A np.ndarray of attribute values, or None if the attribute is None.
        """
        if attribute is None:
            return None

        attribute_vals = []
        is_nested = False
        if self.input_cols:
            iterable_types = [list, tuple, set]
            is_nested = type(attribute[self.input_cols[0]]) in iterable_types

        for input_col in self.input_cols:
            val = attribute[input_col]
            attribute_vals.append(np.array(val) if is_nested else val)
        attribute_arr = np.array(attribute_vals, dtype=object if is_nested else None)

        if dtype:
            attribute_arr = attribute_arr.astype(dtype)

        return attribute_arr

    def _transform_sklearn(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input dataset using the fitted sklearn transform obj.

        Args:
            dataset: Input dataset to transform.

        Returns:
            Transformed dataset

        Raises:
            SnowflakeMLException: If the supplied output columns don't match that of the transformed array.
        """
        self._enforce_fit()
        dataset = dataset.copy()
        sklearn_transform = self.to_sklearn()
        transformed_data = sklearn_transform.transform(dataset[self.input_cols])
        shape = transformed_data.shape
        if (len(shape) == 1 and len(self.output_cols) != 1) or (len(shape) > 1 and shape[1] != len(self.output_cols)):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_ATTRIBUTE,
                original_exception=RuntimeError(
                    modeling_error_messages.SIZE_MISMATCH.format(
                        "output_cols", len(self.output_cols), "transformed array shape", shape
                    )
                ),
            )

        if len(shape) == 1:
            transformed_data = np.reshape(transformed_data, (-1, 1))
        dataset[self.output_cols] = transformed_data
        return dataset

    def _validate_data_has_no_nulls(self, dataset: snowpark.DataFrame) -> None:
        """
        Validate that the supplied data does not contain null values in the supplied input columns.

        Args:
            dataset: DataFrame to validate.

        Raises:
            SnowflakeMLException: If the dataset contains nulls in the input_cols.
        """
        self._check_input_cols()

        null_count_columns = []
        for input_col in self.input_cols:
            col = F.count(F.lit("*")) - F.count(dataset[input_col])
            null_count_columns.append(col)

        null_counts = dataset.agg(*null_count_columns).collect(
            statement_params=telemetry.get_statement_params(PROJECT, SUBPROJECT, self.__class__.__name__)
        )
        assert len(null_counts) == 1

        invalid_columns = {col: n for (col, n) in zip(self.input_cols, null_counts[0].as_dict().values()) if n > 0}

        if any(invalid_columns):
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INVALID_DATA,
                original_exception=ValueError(
                    "Dataset may not contain nulls."
                    f"The following columns have non-zero numbers of nulls: {invalid_columns}."
                ),
            )

    def _drop_input_columns(
        self, dataset: Union[snowpark.DataFrame, pd.DataFrame]
    ) -> Union[snowpark.DataFrame, pd.DataFrame]:
        """
        Drop input column for given dataset.

        Args:
            dataset: The input Dataset. Either a Snowflake DataFrame or Pandas DataFrame.

        Returns:
            Return a dataset with input columns dropped.

        Raises:
            SnowflakeMLException: drop_input_cols flag must be true before calling this function.
        """
        if not self._drop_input_cols:
            raise exceptions.SnowflakeMLException(
                error_code=error_codes.INTERNAL_PYTHON_ERROR,
                original_exception=RuntimeError("drop_input_cols must set true."),
            )

        # In case of input_cols and output_cols are the same, compare with get_output_cols()
        input_subset = list(set(self.input_cols) - set(self.get_output_cols()))

        super()._check_dataset_type(dataset)
        if isinstance(dataset, snowpark.DataFrame):
            return dataset.drop(input_subset)
        else:
            return dataset.drop(columns=input_subset)
